from docutils.parsers.rst import directives
from docutils.nodes import literal_block, Text
from sphinx.directives.code import LiteralInclude

class ExtendedLiteralInclude(LiteralInclude):
    """
    We extend literalinclude to allow :replace: strings to be provided. This helps
    us fix imports that may be confusing in tutorials etc.
    """

    option_spec = LiteralInclude.option_spec.copy()
    option_spec['replace'] = directives.unchanged

    def run(self):
        # Call the original run method to get the literal include content
        original_result = super().run()

        if 'replace' in self.options:
            replace_option = self.options['replace']
            replacements = self.parse_replacements(replace_option)
            if not replacements:
                return original_result

            # Modify the content in original_result
            def make_replacements(node):
                if isinstance(node, literal_block):
                    content = node.rawsource
                    for old_text, new_text in replacements.items():
                        content = content.replace(old_text, new_text)

                    node.rawsource = content
                    node.children[0] = Text(content)
                    node.children[0].parent = node

            def walk_nodes(nodes):
                for node in nodes:
                    make_replacements(node)
                    if hasattr(node, 'children'):
                        walk_nodes(node.children)
            
            walk_nodes(original_result)

        return original_result

    def parse_replacements(self, replace_option):
        replacements = {}
        lines = replace_option.splitlines()
        for line in lines:
            if ':' in line:
                old_text, new_text = line.split(':', 1)
                replacements[old_text.strip()] = new_text.strip()
        return replacements



"""
Tests for the typer-native style interface.
"""

from django.core.management import call_command

from django_typer.management import get_command
from tests.utils import run_command

from . import test_native


class TestNativeInheritance(test_native.TestNative):
    command = "native_inheritance1"


class TestNativeInheritanceWithSelf(test_native.TestNative):
    command = "native_inheritance2"


class TestNativeInheritanceGroups(test_native.TestNativeGroups):
    command = "native_inheritance3"


class TestNativeInheritanceGroupsWithSelf(test_native.TestNativeGroups):
    command = "native_inheritance4"


native_override_init_help_rich = """
 Usage: ./manage.py native_override_init [OPTIONS] COMMAND [ARGS]...            
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --fog     --no-fog      [default: no-fog]                                    │
│ --help                  Show this message and exit.                          │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --verbosity        INTEGER RANGE [0<=x<=3]  Verbosity level; 0=minimal       │
│                                             output, 1=normal output,         │
│                                             2=verbose output, 3=very verbose │
│                                             output                           │
│                                             [default: 0]                     │
│ --version                                   Show program's version number    │
│                                             and exit.                        │
│ --settings         TEXT                     The Python path to a settings    │
│                                             module, e.g.                     │
│                                             "myproject.settings.main". If    │
│                                             this isn't provided, the         │
│                                             DJANGO_SETTINGS_MODULE           │
│                                             environment variable will be     │
│                                             used.                            │
│ --no-color                                  Don't colorize the command       │
│                                             output.                          │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ grp1   Override GRP1 (initialize only)                                       │
│ main                                                                         │
╰──────────────────────────────────────────────────────────────────────────────╯
"""


native_override_init_grp1_help_rich = """
 Usage: ./manage.py native_override_init grp1 [OPTIONS] COMMAND [ARGS]...       
                                                                                
 Override GRP1 (initialize only)                                                
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --flag        INTEGER  [default: 4]                                          │
│ --help                 Show this message and exit.                           │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ cmd1                                                                         │
│ cmd2                                                                         │
│ subgrp                                                                       │
╰──────────────────────────────────────────────────────────────────────────────╯
"""


class TestNativeInitOverride(test_native.TestNativeGroups):
    command = "native_override_init"

    commands = [
        ("{command}", native_override_init_help_rich),
        ("{command} main", test_native.native_groups_main_help_rich),
        ("{command} grp1", native_override_init_grp1_help_rich),
        ("{command} grp1 subgrp", test_native.native_groups_grp1_subgrp_help_rich),
        ("{command} grp1 cmd1", test_native.native_groups_grp1_cmd1_help_rich),
        ("{command} grp1 cmd2", test_native.native_groups_grp1_cmd2_help_rich),
    ]

    def test_native_groups_direct(self):
        native_groups = get_command(self.command)
        native_groups.init(verbosity=3)
        self.assertEqual(native_groups.main("Brian"), {"verbosity": 3, "name": "Brian"})

        native_groups.init_grp1(flag=8)

        self.assertEqual(native_groups.cmd1(5), {"verbosity": 3, "flag": 8, "count": 5})
        native_groups.cmd1(5)

        native_groups.init_grp1()
        self.assertEqual(
            native_groups.cmd2(3.5), {"verbosity": 3, "flag": 4, "fraction": 3.5}
        )
        self.assertEqual(
            native_groups.init(fog=True, verbosity=7), {"verbosity": 7, "fog": True}
        )

    def test_native_groups_run(self):
        self.assertEqual(
            run_command(
                self.command, *self.settings, "--verbosity", "3", "main", "Brian"
            )[0].strip(),
            str({"verbosity": 3, "name": "Brian"}),
        )

        self.assertEqual(
            run_command(
                self.command,
                *self.settings,
                "--verbosity",
                "2",
                "grp1",
                "--flag",
                "3",
                "cmd1",
                "5",
            )[0].strip(),
            str({"verbosity": 2, "flag": 3, "count": 5}),
        )

        self.assertEqual(
            run_command(
                self.command, *self.settings, "--verbosity", "1", "grp1", "cmd2", "2.5"
            )[0].strip(),
            str({"verbosity": 1, "flag": 4, "fraction": 2.5}),
        )

    def test_native_groups_call(self):
        self.assertEqual(
            call_command(self.command, "main", "Brian", verbosity=3),
            {"verbosity": 3, "name": "Brian"},
        )

        self.assertEqual(
            call_command(self.command, "grp1", "cmd1", "5", verbosity=2, flag=9),
            {"verbosity": 2, "flag": 9, "count": 5},
        )

        self.assertEqual(
            call_command(
                self.command, "--verbosity", "1", "grp1", "--flag", "2", "cmd2", "2.5"
            ),
            {"verbosity": 1, "flag": 2, "fraction": 2.5},
        )

    def test_native_groups_direct_run_subgrp(self, flag=3):
        return super().test_native_groups_direct_run_subgrp(flag=flag)

    def test_native_groups_call_subgrp(self, flag=6):
        super().test_native_groups_call_subgrp(flag=flag)

    def test_native_groups_run_subgrp(self, flag=4):
        super().test_native_groups_run_subgrp(flag=flag)


native_tweaks_help_rich = """
 Usage: ./manage.py native_tweaks [OPTIONS] NAME                                
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    name      TEXT  [required]                                              │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Django ─────────────────────────────────────────────────────────────────────╮
│ --version                                    Show program's version number   │
│                                              and exit.                       │
│ --verbosity         INTEGER RANGE [0<=x<=3]  Verbosity level; 0=minimal      │
│                                              output, 1=normal output,        │
│                                              2=verbose output, 3=very        │
│                                              verbose output                  │
│                                              [default: 1]                    │
│ --settings          TEXT                     The Python path to a settings   │
│                                              module, e.g.                    │
│                                              "myproject.settings.main". If   │
│                                              this isn't provided, the        │
│                                              DJANGO_SETTINGS_MODULE          │
│                                              environment variable will be    │
│                                              used.                           │
│ --pythonpath        PATH                     A directory to add to the       │
│                                              Python path, e.g.               │
│                                              "/home/djangoprojects/myprojec… │
│ --no-color                                   Don't colorize the command      │
│                                              output.                         │
╰──────────────────────────────────────────────────────────────────────────────╯
"""


native_override_grp1_help_rich = """
 Usage: ./manage.py native_override grp1 [OPTIONS] COMMAND [ARGS]...            
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --flag    --no-flag      [default: no-flag]                                  │
│ --help                   Show this message and exit.                         │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ cmd1                                                                         │
│ cmd2                                                                         │
│ subgrp   Override SUBGROUP                                                   │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

native_override_subgrp_help_rich = """
 Usage: ./manage.py native_override grp1 subgrp [OPTIONS] COMMAND [ARGS]...     
                                                                                
 Override SUBGROUP                                                              
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --option1    --no-option1      [default: no-option1]                         │
│ --option2    --no-option2      [default: no-option2]                            │
│ --help                         Show this message and exit.                   │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Commands ───────────────────────────────────────────────────────────────────╮
│ sg-cmd1   Subgroup command 1. No args.                                       │
│ sg-cmd2   Subgroup command 2, Takes an int.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

native_override_sgcmd1_help_rich = """
 Usage: ./manage.py native_override grp1 subgrp sg-cmd1 [OPTIONS]               
                                                                                
 Subgroup command 1. No args.                                                   
                                                                                
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""

native_override_sgcmd2_help_rich = """
 Usage: ./manage.py native_override grp1 subgrp sg-cmd2 [OPTIONS] NUMBER        
                                                                                
 Subgroup command 2, Takes an int.                                              
                                                                                
╭─ Arguments ──────────────────────────────────────────────────────────────────╮
│ *    number      INTEGER  [required]                                         │
╰──────────────────────────────────────────────────────────────────────────────╯
╭─ Options ────────────────────────────────────────────────────────────────────╮
│ --help          Show this message and exit.                                  │
╰──────────────────────────────────────────────────────────────────────────────╯
"""


class TestNativeOverrides(test_native.TestNativeGroups):
    command = "native_override"

    commands = [
        ("{command}", test_native.native_groups_help_rich),
        ("{command} main", test_native.native_groups_main_help_rich),
        ("{command} grp1", native_override_grp1_help_rich),
        ("{command} grp1 subgrp", native_override_subgrp_help_rich),
        ("{command} grp1 subgrp sg-cmd1", native_override_sgcmd1_help_rich),
        ("{command} grp1 subgrp sg-cmd2", native_override_sgcmd2_help_rich),
        ("{command} grp1 cmd1", test_native.native_groups_grp1_cmd1_help_rich),
        ("{command} grp1 cmd2", test_native.native_groups_grp1_cmd2_help_rich),
    ]

    def test_native_groups_direct_run_subgrp(self):
        native_groups = get_command(self.command)
        native_groups.init(verbosity=3)
        native_groups.init_grp1(flag=True)
        native_groups.subgrp(option2=False, option1=True)
        self.assertEqual(
            native_groups.sg_cmd1(),
            {"verbosity": 3, "flag": True, "option1": True, "option2": False},
        )
        self.assertEqual(
            native_groups.sg_cmd2(7),
            {
                "verbosity": 3,
                "flag": True,
                "option1": True,
                "option2": False,
                "number": 7,
            },
        )

    def test_native_groups_run_subgrp(self):
        self.assertEqual(
            run_command(
                self.command,
                *self.settings,
                "--verbosity",
                "1",
                "grp1",
                "subgrp",
                "--option1",
                "sg-cmd1",
            )[0].strip(),
            str({"verbosity": 1, "flag": False, "option1": True, "option2": False}),
        )

        self.assertEqual(
            run_command(
                self.command,
                *self.settings,
                "--verbosity",
                "2",
                "grp1",
                "--flag",
                "subgrp",
                "--option2",
                "sg-cmd2",
                "5",
            )[0].strip(),
            str(
                {
                    "verbosity": 2,
                    "flag": True,
                    "option1": False,
                    "option2": True,
                    "number": 5,
                }
            ),
        )

    def test_native_groups_call_subgrp(self):
        self.assertEqual(
            call_command(
                self.command,
                "--verbosity",
                "3",
                "grp1",
                "subgrp",
                "--no-option2",
                "sg-cmd1",
                option1=True,
                flag=True,
            ),
            {"verbosity": 3, "flag": True, "option1": True, "option2": False},
        )

        self.assertEqual(
            call_command(
                self.command,
                "grp1",
                "--flag",
                "subgrp",
                "--option2",
                "sg-cmd2",
                "5",
                verbosity=2,
            ),
            {
                "verbosity": 2,
                "flag": True,
                "option1": False,
                "option2": True,
                "number": 5,
            },
        )



# Generated by Django 5.1.7 on 2025-03-31 19:18

import tests.apps.dj5plus.models
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="ChoicesShellCompleteTesterDJ5Plus",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "char_choice",
                    models.CharField(
                        choices=[
                            ("git", "Git"),
                            ("svn", "Subversion (SVN)"),
                            ("hg", "Mercurial (Hg)"),
                            ("bzr", "Bazaar (Bzr)"),
                            ("cvs", "CVS"),
                            ("perforce", "Perforce"),
                            ("fossil", "Fossil"),
                            ("darcs", "Darcs"),
                            ("monotone", "Monotone"),
                        ],
                        db_index=True,
                        default=None,
                        max_length=12,
                        null=True,
                    ),
                ),
                (
                    "int_choice",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (1, "One"),
                            (2, "Two"),
                            (3, "Mercurial (Hg)"),
                            (4, "Bazaar (Bzr)"),
                            (5, "CVS"),
                            (6, "Perforce"),
                            (7, "Fossil"),
                            (8, "Darcs"),
                            (9, "Monotone"),
                            (10, "Ten"),
                            (11, "Eleven"),
                            (12, "Twelve"),
                        ],
                        db_index=True,
                        default=None,
                        null=True,
                    ),
                ),
                (
                    "ip_choice",
                    models.GenericIPAddressField(
                        choices=tests.apps.dj5plus.models.get_ip_choices,
                        db_index=True,
                        default=None,
                        null=True,
                    ),
                ),
            ],
        ),
    ]



from .finalize_simple import Command as FinalizeSimple


class Command(FinalizeSimple):
    @FinalizeSimple.finalize()
    def collect(self, result, **_):
        assert isinstance(self, Command)
        assert result == "handle"
        return "collect: {}".format(result)



from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand):
    @initialize()
    def init(self, common_option: bool = False):
        # you can define common options that will be available to all
        # subcommands of the command, and implement common initialization
        # logic here. This will be invoked before the chosen command
        self.common_option = common_option

    @command()
    def subcommand1(self):
        return self.common_option

    @command()
    def subcommand2(self):
        return self.common_option



import tarfile
import typing as t
from pathlib import Path

import typer
from django.conf import settings

from django_typer.management import Typer
from tests.apps.examples.plugins.backup.management.commands import (
    backup_typer,
)

# to inherit all commands defined upstream pass the root typer app from upstream
# to our root app here. (This is not a standard typer interface.)
app = Typer(backup_typer.app)


@app.command()
def media(
    self,
    filename: t.Annotated[
        str,
        typer.Option(
            "-f",
            "--filename",
            help=("The name of the file to use for the media backup tar."),
        ),
    ] = "media.tar.gz",
):
    """
    Backup the media files (i.e. those files in MEDIA_ROOT).
    """
    media_root = Path(settings.MEDIA_ROOT)
    output_file = self.output_directory / filename

    # backup the media directory into the output file as a gzipped tar
    typer.echo(f"Backing up {media_root} to {output_file}")
    with tarfile.open(output_file, "w:gz") as tar:
        tar.add(media_root, arcname=media_root.name)



import contextlib
from io import StringIO

from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command, rich_installed


class TestFinalize(TestCase):
    def test_finalize_multi_kwargs_run(
        self, command="finalize_multi_kwargs", show_locals=rich_installed
    ):
        stdout, _, _ = run_command(command, "cmd1")
        self.assertEqual(
            stdout.strip(),
            "finalized: ['cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False, 'show_locals': None}"
            if show_locals
            else "finalized: ['cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False}",
        )
        stdout, _, _ = run_command(command, "cmd2", "3", "cmd1")
        self.assertEqual(
            stdout.strip(),
            "finalized: ['cmd2 3', 'cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False, 'show_locals': None}"
            if show_locals
            else "finalized: ['cmd2 3', 'cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False}",
        )

    def test_finalize_multi_named_param_run(self):
        self.test_finalize_multi_kwargs_run(
            command="finalize_multi_named_param", show_locals=True
        )

    def test_finalize_no_params_run(self):
        stdout, _, _ = run_command("finalize_multi_no_params", "cmd1")
        self.assertEqual(
            stdout.strip(),
            "finalized: ['cmd1 1']",
        )

        stdout, _, _ = run_command("finalize_multi_no_params", "cmd2", "3", "cmd1")
        self.assertEqual(
            stdout.strip(),
            "finalized: ['cmd2 3', 'cmd1 1']",
        )

    def test_finalize_multi_kwargs_call(
        self, command="finalize_multi_kwargs", show_locals=rich_installed
    ):
        # todo - excluded common options should not appear?
        call_command(command, "cmd1")
        with contextlib.redirect_stdout(StringIO()) as out:
            call_command(command, "cmd1")
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False, 'show_locals': None}"
                if show_locals
                else "finalized: ['cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False}",
            )

            out.truncate(0)
            out.seek(0)

            call_command(command, "cmd2", "5", "cmd1")
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd2 5', 'cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False, 'show_locals': None}"
                if show_locals
                else "finalized: ['cmd2 5', 'cmd1 1'] | {'force_color': False, 'no_color': False, 'traceback': False}",
            )

            out.truncate(0)
            out.seek(0)

            call_command(
                command,
                "--traceback",
                "cmd2",
                "3",
                "cmd1",
                "--arg1",
                "2",
            )
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd2 3', 'cmd1 2'] | {'force_color': False, 'no_color': False, 'traceback': True, 'show_locals': None}"
                if show_locals
                else "finalized: ['cmd2 3', 'cmd1 2'] | {'force_color': False, 'no_color': False, 'traceback': True}",
            )

    def test_finalize_multi_named_param_call(self):
        self.test_finalize_multi_kwargs_call(
            command="finalize_multi_named_param", show_locals=True
        )

    def test_finalize_multi_no_params(self):
        # todo - excluded common options should not appear?
        call_command("finalize_multi_no_params", "cmd1")
        with contextlib.redirect_stdout(StringIO()) as out:
            call_command("finalize_multi_no_params", "cmd1")
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd1 1']",
            )

            out.truncate(0)
            out.seek(0)

            call_command("finalize_multi_no_params", "cmd2", "5", "cmd1")
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd2 5', 'cmd1 1']",
            )

            out.truncate(0)
            out.seek(0)

            call_command(
                "finalize_multi_no_params",
                "--traceback",
                "cmd2",
                "3",
                "cmd1",
                "--arg1",
                "2",
            )
            self.assertEqual(
                out.getvalue().strip(),
                "finalized: ['cmd2 3', 'cmd1 2']",
            )

    def test_finalize_multi_kwargs_obj(
        self, command="finalize_multi_kwargs", kwargs=""
    ):
        from tests.apps.test_app.management.commands.finalize_multi_kwargs import (
            Command,
        )

        out = StringIO()
        finalize4 = get_command(command, Command, stdout=out)
        self.assertEqual(
            finalize4.final([finalize4.cmd1(3), finalize4.cmd2(2)]).strip(),
            f"finalized: ['cmd1 3', 'cmd2 2'] | {{{kwargs}}}",
        )

    def test_finalize_multi_named_param_obj(self):
        self.test_finalize_multi_kwargs_obj(
            command="finalize_multi_named_param",
            kwargs="'force_color': None, 'no_color': None, 'traceback': None, 'show_locals': None",
        )

    def test_finalize_multi_no_params_obj(
        self, command="finalize_multi_no_params", kwargs=""
    ):
        from tests.apps.test_app.management.commands.finalize_multi_no_params import (
            Command,
        )

        out = StringIO()
        finalize_multi_no_params = get_command(command, Command, stdout=out)
        self.assertEqual(
            finalize_multi_no_params.final(
                [finalize_multi_no_params.cmd1(3), finalize_multi_no_params.cmd2(2)]
            ).strip(),
            f"finalized: ['cmd1 3', 'cmd2 2']",
        )

    def test_finalize_simple_run(self):
        stdout, _, _ = run_command("finalize_simple")
        self.assertEqual(stdout.strip(), "finalized: handle")

    def test_finalize_simple_call(self):
        out = StringIO()
        call_command("finalize_simple", stdout=out)
        self.assertEqual(out.getvalue().strip(), "finalized: handle")

    def test_finalize_simple_obj(self):
        from tests.apps.test_app.management.commands.finalize_simple import Command

        out = StringIO()
        finalize_simple = get_command("finalize_simple", Command, stdout=out)
        self.assertEqual(
            finalize_simple.final(finalize_simple()).strip(),
            "finalized: handle",
        )

    def test_finalize_static_run(self):
        stdout, _, _ = run_command("finalize_static")
        self.assertEqual(stdout.strip(), "finalized: handle | traceback=False")

        stdout, _, _ = run_command("finalize_static", "--traceback")
        self.assertEqual(stdout.strip(), "finalized: handle | traceback=True")

    def test_finalize_static_call(self):
        out = StringIO()
        call_command("finalize_static", stdout=out)
        self.assertEqual(out.getvalue().strip(), "finalized: handle | traceback=False")

        out = StringIO()
        call_command("finalize_static", traceback=True, stdout=out)
        self.assertEqual(out.getvalue().strip(), "finalized: handle | traceback=True")

    def test_finalize_static_obj(self):
        from tests.apps.test_app.management.commands.finalize_static import Command

        out = StringIO()
        finalize_static = get_command("finalize_static", Command, stdout=out)
        self.assertEqual(
            finalize_static.final(finalize_static()).strip(),
            "finalized: handle | traceback=False",
        )

    def test_finalize_subgroups_run(self):
        stdout, _, _ = run_command("finalize_subgroups", "grp2", "cmd4", "cmd3")
        self.assertEqual(
            stdout.strip(),
            "root_final: grp2_final: ['cmd4', 'cmd3'] | g2_opt=True | init_opt=False",
        )

        stdout, _, _ = run_command(
            "finalize_subgroups", "--init-opt", "grp2", "--no-g2-opt", "cmd3"
        )
        self.assertEqual(
            stdout.strip(),
            "root_final: grp2_final: ['cmd3'] | g2_opt=False | init_opt=True",
        )

        stdout, _, _ = run_command(
            "finalize_subgroups", "--init-opt", "grp1", "--no-g1-opt", "cmd2", "cmd2"
        )
        self.assertEqual(
            stdout.strip(),
            "root_final: grp1_final: ['cmd2', 'cmd2'] | g1_opt=False | init_opt=True",
        )

        stdout, _, _ = run_command("finalize_subgroups", "grp1", "cmd2", "cmd1")
        self.assertEqual(
            stdout.strip(),
            "root_final: grp1_final: ['cmd2', 'cmd1'] | g1_opt=False | init_opt=False",
        )

    def test_finalize_subgroups_call(self):
        out = StringIO()
        call_command(
            "finalize_subgroups", "grp2", "cmd4", "cmd3", stdout=out, call_command=True
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp2_final: ['cmd4', 'cmd3'] | g2_opt=True | init_opt=False",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups",
            "grp2",
            "cmd3",
            init_opt=True,
            g2_opt=False,
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp2_final: ['cmd3'] | g2_opt=False | init_opt=True",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups",
            "grp1",
            "--no-g1-opt",
            "cmd2",
            "cmd2",
            init_opt=True,
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp1_final: ['cmd2', 'cmd2'] | g1_opt=False | init_opt=True",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups", "grp1", "cmd2", "cmd1", stdout=out, call_command=True
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp1_final: ['cmd2', 'cmd1'] | g1_opt=False | init_opt=False",
        )

    def test_finalize_subgroups_obj(self):
        from tests.apps.test_app.management.commands.finalize_subgroups import Command

        out = StringIO()
        finalize_subgroups = get_command("finalize_subgroups", Command, stdout=out)
        finalize_subgroups.init(init_opt=True, direct_call=True)
        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp1_final(
                        [
                            finalize_subgroups.grp1.cmd2(),
                            finalize_subgroups.grp1.cmd2(),
                        ]
                    )
                ],
                init_opt=True,
            ).strip(),
            "root_final: [\"grp1_final: ['cmd2', 'cmd2'] | g1_opt=None\"] | init_opt=True",
        )

        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp1.grp1_final(
                        [
                            finalize_subgroups.cmd1(),
                            finalize_subgroups.cmd2(),
                        ]
                    )
                ],
                init_opt=False,
            ).strip(),
            "root_final: [\"grp1_final: ['cmd1', 'cmd2'] | g1_opt=None\"] | init_opt=False",
        )
        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp2.grp2_final(
                        [
                            finalize_subgroups.cmd3(),
                            finalize_subgroups.cmd4(),
                        ],
                        g2_opt=False,
                    )
                ],
                init_opt=False,
            ).strip(),
            "root_final: [\"grp2_final: ['cmd3', 'cmd4'] | g2_opt=False\"] | init_opt=False",
        )

    def test_finalize_override_run(self):
        stdout, _, _ = run_command("finalize_override")
        self.assertEqual(stdout.strip(), "collect: handle")

    def test_finalize_override_call(self):
        out = StringIO()
        call_command("finalize_override", stdout=out)
        self.assertEqual(out.getvalue().strip(), "collect: handle")

    def test_finalize_override_obj(self):
        from tests.apps.test_app.management.commands.finalize_override import Command

        out = StringIO()
        finalize_override = get_command("finalize_override", Command, stdout=out)
        self.assertEqual(
            finalize_override.collect(finalize_override()).strip(),
            "collect: handle",
        )
        self.assertEqual(
            finalize_override.final(finalize_override()).strip(),
            "finalized: handle",
        )

    def test_finalize_override_inherit_run(self):
        stdout, _, _ = run_command("finalize_override_inherit")
        self.assertEqual(stdout.strip(), "ext_collect: handle")

    def test_finalize_override_inherit_call(self):
        out = StringIO()
        call_command("finalize_override_inherit", stdout=out)
        self.assertEqual(out.getvalue().strip(), "ext_collect: handle")

    def test_finalize_override_inherit_obj(self):
        from tests.apps.test_app.management.commands.finalize_override_inherit import (
            Command,
        )

        out = StringIO()
        finalize_override = get_command(
            "finalize_override_inherit", Command, stdout=out
        )
        self.assertEqual(
            finalize_override.ext_collect(finalize_override()).strip(),
            "ext_collect: handle",
        )
        self.assertEqual(
            finalize_override.final(finalize_override()).strip(),
            "finalized: handle",
        )

    def test_finalize_subgroups_inherit_run(self):
        stdout, _, _ = run_command("finalize_subgroups_inherit", "grp2", "cmd4", "cmd3")
        self.assertEqual(
            stdout.strip(),
            "root_final: grp2_collect: grp2_final: ['cmd4', 'cmd3'] | g2_opt=True | init_opt=False",
        )

        stdout, _, _ = run_command(
            "finalize_subgroups_inherit", "--init-opt", "grp2", "--no-g2-opt", "cmd3"
        )
        self.assertEqual(
            stdout.strip(),
            "root_final: grp2_collect: grp2_final: ['cmd3'] | g2_opt=False | init_opt=True",
        )

        stdout, _, _ = run_command(
            "finalize_subgroups_inherit",
            "--init-opt",
            "grp1",
            "--no-g1-opt",
            "cmd2",
            "cmd2",
        )
        self.assertEqual(
            stdout.strip(),
            "root_final: grp1_collect: grp1_final: ['cmd2', 'cmd2'] | g1_opt=False | init_opt=True",
        )

        stdout, _, _ = run_command("finalize_subgroups_inherit", "grp1", "cmd2", "cmd1")
        self.assertEqual(
            stdout.strip(),
            "root_final: grp1_collect: grp1_final: ['cmd2', 'cmd1'] | g1_opt=False | init_opt=False",
        )

    def test_finalize_subgroups_inherit_call(self):
        out = StringIO()
        call_command(
            "finalize_subgroups_inherit",
            "grp2",
            "cmd4",
            "cmd3",
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp2_collect: grp2_final: ['cmd4', 'cmd3'] | g2_opt=True | init_opt=False",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups_inherit",
            "grp2",
            "cmd3",
            init_opt=True,
            g2_opt=False,
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp2_collect: grp2_final: ['cmd3'] | g2_opt=False | init_opt=True",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups_inherit",
            "grp1",
            "--no-g1-opt",
            "cmd2",
            "cmd2",
            init_opt=True,
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp1_collect: grp1_final: ['cmd2', 'cmd2'] | g1_opt=False | init_opt=True",
        )

        out = StringIO()
        call_command(
            "finalize_subgroups_inherit",
            "grp1",
            "cmd2",
            "cmd1",
            stdout=out,
            call_command=True,
        )
        self.assertEqual(
            out.getvalue().strip(),
            "root_final: grp1_collect: grp1_final: ['cmd2', 'cmd1'] | g1_opt=False | init_opt=False",
        )

    def test_finalize_subgroups_inherit_obj(self):
        from tests.apps.test_app.management.commands.finalize_subgroups_inherit import (
            Command,
        )

        out = StringIO()
        finalize_subgroups = get_command(
            "finalize_subgroups_inherit", Command, stdout=out
        )
        finalize_subgroups.init(init_opt=True, direct_call=True)
        self.assertEqual(finalize_subgroups.cmd5(), "cmd5")
        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp1_collect(
                        [
                            finalize_subgroups.grp1.cmd2(),
                            finalize_subgroups.grp1.cmd2(),
                        ]
                    )
                ],
                init_opt=True,
            ).strip(),
            "root_final: [\"grp1_collect: grp1_final: ['cmd2', 'cmd2'] | g1_opt=None\"] | init_opt=True",
        )

        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp1.grp1_collect(
                        [
                            finalize_subgroups.cmd1(),
                            finalize_subgroups.cmd2(),
                        ]
                    )
                ],
                init_opt=False,
            ).strip(),
            "root_final: [\"grp1_collect: grp1_final: ['cmd1', 'cmd2'] | g1_opt=None\"] | init_opt=False",
        )
        self.assertEqual(
            finalize_subgroups.root_final(
                [
                    finalize_subgroups.grp2.grp2_final(
                        [
                            finalize_subgroups.cmd3(),
                            finalize_subgroups.cmd4(),
                        ],
                        g2_opt=False,
                    )
                ],
                init_opt=False,
            ).strip(),
            "root_final: [\"grp2_collect: grp2_final: ['cmd3', 'cmd4'] | g2_opt=False\"] | init_opt=False",
        )



import typing as t

from django.core.management.base import CommandError
from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import TyperCommand
from tests.apps.examples.polls.models import Question as Poll


class Command(TyperCommand):
    help = "Closes the specified poll for voting"

    def handle(
        self,
        poll_ids: t.Annotated[
            t.List[int],
            Argument(help=_("The database IDs of the poll(s) to close.")),
        ],
        delete: t.Annotated[
            bool, Option(help=_("Delete poll instead of closing it."))
        ] = False,
    ):
        for poll_id in poll_ids:
            try:
                poll = Poll.objects.get(pk=poll_id)
            except Poll.DoesNotExist:
                raise CommandError(f'Poll "{poll_id}" does not exist')

            poll.opened = False
            poll.save()

            self.stdout.write(
                self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
            )

            if delete:
                poll.delete()



import typing as t
import typer
from click import Context, Parameter
from click.shell_completion import CompletionItem
from django.apps import apps

from django_typer.management import TyperCommand, command


# the completer function signature must match this exactly
def complete_app_label(
    ctx: Context, param: Parameter, incomplete: str
) -> t.List[CompletionItem]:
    # don't offer apps that are already present as completion suggestions
    present = [app.label for app in (ctx.params.get(param.name or "") or [])]
    return [
        CompletionItem(app.label)
        for app in apps.get_app_configs()
        if app.label.startswith(incomplete) and app.label not in present
    ]


class Command(TyperCommand):
    @command()
    def handle(
        self,
        apps: t.Annotated[
            t.List[str],
            typer.Argument(
                help="The app label",
                # pass the completer function here
                shell_complete=complete_app_label,
            ),
        ],
    ):
        print(apps)



import importlib
import inspect
import typing as t

import typer.core
from django.apps import AppConfig, apps
from django.core.management.base import CommandError

from django_typer.management import Typer, TyperCommand, get_command
from django_typer import completers, parsers, utils

from pathlib import Path

import graphviz
import typer


try:
    from enum import StrEnum

    class Format(StrEnum):
        png = "png"
        svg = "svg"
        pdf = "pdf"
        dot = "dot"

except ImportError:
    from enum import Enum

    class Format(str, Enum):
        png = "png"
        svg = "svg"
        pdf = "pdf"
        dot = "dot"

        def __str__(self):
            return self.value


class Command(TyperCommand):
    """
    Graph the Typer app tree associated with the given command.
    """

    cmd: TyperCommand
    cmd_name: str
    app_name: str
    plugin: t.Optional[str] = None
    dot = graphviz.Digraph()
    level: int = -1
    show_ids: bool = False
    seen = set()

    def handle(
        self,
        commands: t.Annotated[
            t.List[str],
            typer.Argument(
                help="Import path(s) to the command to graph, or simply the name of the command.",
                shell_complete=completers.chain(
                    completers.complete_import_path,
                    completers.commands(allow_duplicates=True),
                    allow_duplicates=True,
                ),
            ),
        ],
        output: t.Annotated[
            Path,
            typer.Option(
                "-o",
                "--output",
                help="The path to save the graph to.",
                shell_complete=completers.complete_path,
            ),
        ] = Path("{command}_app_tree"),
        format: t.Annotated[
            Format,
            typer.Option(
                "-f",
                "--format",
                help="The format to save the graph in.",
                shell_complete=completers.these_strings(list(Format)),
            ),
        ] = Format.png,
        show_ids: t.Annotated[
            bool, typer.Option(help="Show the object identifiers.")
        ] = show_ids,
        instantiate: t.Annotated[
            bool,
            typer.Option(help="Instantiate the command before graphing the app tree."),
        ] = True,
        load_order: t.Annotated[
            t.List[AppConfig],
            typer.Option(
                "-l",
                "--load",
                parser=parsers.parse_app_label,
                shell_complete=completers.complete_app_label,
                help="Load plugins from these apps after each command is graphed in the order provided.",
            ),
        ] = [],
    ):
        assert len(load_order) <= len(commands), (
            "Cannot provide more load options than commands"
        )

        plugin_loads = []

        for app in load_order:
            to_load = {}
            for cmd, plugins in list(utils._command_plugins.items()):
                to_load[cmd] = []
                for plugin in list(plugins):
                    if (
                        self.get_app_from_module(plugin.__name__.split(".")[0:-2])
                        == app
                    ):
                        to_load[cmd].append(plugin)
            plugin_loads.append(to_load)

        if load_order:
            utils._command_plugins = {}

        self.show_ids = show_ids
        for command in commands:
            self.level = -1
            self.cmd_name = command.split(".")[-1]
            if "." in command:
                self.cmd = getattr(importlib.import_module(command), "Command")
                if instantiate:
                    self.cmd = self.cmd()
            elif instantiate:
                self.cmd = get_command(command, TyperCommand)
            else:
                raise CommandError("Cannot instantiate a command that is not imported.")

            self.app_name = self.get_app_from_module(
                self.cmd.__module__.split(".")[0:-3]
            ).name

            output = Path(output.parent) / Path(
                output.name.format(command=self.cmd_name)
            )

            self.visit_app(self.cmd.typer_app)

            of = output.parent / f"{output.stem}.{format}"
            num = 1
            while of.exists():
                of = output.parent / f"{output.stem}({num}).{format}"
                num += 1
            self.dot.render(of.parent / of.stem, format=format, view=len(commands) == 1)
            if load_order:
                utils._command_plugins = plugin_loads.pop(0)
                self.plugin = load_order.pop(0).name.split(".")[-1]
            else:
                self.plugin = None

    def get_app_from_module(self, module: t.List[str]) -> t.Optional[AppConfig]:
        app = importlib.import_module(f"{'.'.join(module)}.apps")
        for _, attr in vars(app).items():
            if (
                inspect.isclass(attr)
                and issubclass(attr, AppConfig)
                and attr is not AppConfig
            ):
                return apps.get_app_config(getattr(attr, "label", module[-1]))

    def get_node_name(self, obj: t.Union[typer.models.CommandInfo, Typer]):
        assert obj.callback
        name = (
            obj.callback.__name__
            if isinstance(obj, typer.models.CommandInfo)
            else str(obj.name)
        )
        cli_name = (
            obj.name or name
            if isinstance(obj, typer.models.CommandInfo)
            else obj.info.name or name
        )
        if name != cli_name:
            name += f"({cli_name})"
        if self.show_ids:
            return f"[{hex(id(obj))[-4:]}] {name}"
        return name

    def visit_app(self, app: Typer):
        """
        Walk the typer app tree.
        """
        self.level += 1
        if self.level:
            parent_node = self.get_node_name(app)
        else:
            plugin = f": {self.plugin}" if self.plugin else ""
            parent_node = f"[{self.app_name.split('.')[-1]}{plugin}] {self.cmd_name}"

        style = {}
        if self.level:
            style = {"style": "filled", "fillcolor": "lightblue"}
        self.dot.node(str(id(app)), parent_node, **style)
        for cmd in app.registered_commands:
            node_name = self.get_node_name(cmd)
            self.dot.node(str(id(cmd)), node_name, style="filled", fillcolor="green")
            edge = (str(id(app)), str(id(cmd)))
            if edge not in self.seen:
                self.seen.add(edge)
                self.dot.edge(*edge)
        for grp in app.registered_groups:
            assert grp.typer_instance
            node_name = self.get_node_name(t.cast(Typer, grp.typer_instance))
            edge = str(id(app)), str(id(grp.typer_instance))
            if edge not in self.seen:
                self.seen.add(edge)
                self.dot.edge(*edge)
            self.visit_app(t.cast(Typer, grp.typer_instance))



from django_typer.management import Typer

app = Typer()


# initializers are called callbacks in Typer, but we may also use initialize()
# as an alias
@app.callback()
def init(self, common_option: bool = False):
    # you can define common options that will be available to all subcommands
    # of the command, and implement common initialization logic here. This
    # will be invoked before the chosen command
    self.common_option = common_option


@app.command()
def subcommand1(self):
    return self.common_option


@app.command()
def subcommand2(self):
    return self.common_option



import os
from pathlib import Path

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent


# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "django-insecure-$druy$m$nio-bkagw_%=@(1w)q0=k^mk_5sfk3zi9#4v!%mh*u"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []


# Application definition

INSTALLED_APPS = ["tests.apps.util", "autotranslate", "django_typer"]


# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = "en-us"
# LANGUAGE_CODE = 'de'

AUTOTRANSLATE_TRANSLATOR_SERVICE = "autotranslate.services.GoogleAPITranslatorService"
GOOGLE_TRANSLATE_KEY = (
    (Path(__file__).parent.parent / "apps/util/google_translate.key")
    .read_text()
    .strip()
)
LOCALE_PATHS = [Path(__file__).parent.parent.parent / "django_typer" / "locale"]



import pytest
from django.test import TestCase

from django_typer.management import get_command
from django_typer.management.commands.shellcompletion import Command as ShellComplete
from django_typer.utils import get_win_shell
from shellingham import ShellDetectionFailure


class TestShellResolution(TestCase):
    """
    You must ensure the calling shell is the expected shell for each of these tests to pass.
    """

    def test_powershell(self):
        command = get_command("shellcompletion", ShellComplete)
        command.init()
        self.assertEqual(command.shell, "powershell")

    def test_pwsh(self):
        command = get_command("shellcompletion", ShellComplete)
        command.init()
        self.assertEqual(command.shell, "pwsh")

    def test_bash(self):
        command = get_command("shellcompletion", ShellComplete)
        command.init()
        self.assertEqual(command.shell, "bash")

    def test_zsh(self):
        command = get_command("shellcompletion", ShellComplete)
        command.init()
        self.assertEqual(command.shell, "zsh")

    def test_fish(self):
        command = get_command("shellcompletion", ShellComplete)
        command.init()
        self.assertEqual(command.shell, "fish")


# only run this test if not a subprocess of powershell/pwsh
def test_get_win_shell_failure():
    with pytest.raises(ShellDetectionFailure):
        get_win_shell()



from tests.apps.test_app.management.commands.finalize_override_inherit import (
    Command as FinalizeOverrideInherit,
)


@FinalizeOverrideInherit.finalize()
def ext_collect(self, result, **_):
    assert isinstance(self, FinalizeOverrideInherit)
    assert result == "handle"
    return "ext_collect: {}".format(result)



from django.apps import AppConfig
from django.conf import settings
from django.core.checks import Error, register
from django_typer.utils import register_command_plugins


@register()
def test_check(app_configs, **kwargs):
    errors = []
    if getattr(settings, "DJANGO_TYPER_FAIL_CHECK", False):
        errors.append(
            Error(
                "Test check error",
                hint="Error thrown because settings_fail_check was used.",
                obj=settings,
                id="test_app.E001",
            )
        )
    return errors


class TestAppConfig(AppConfig):
    name = "tests.apps.test_app"
    label = "test_app"
    verbose_name = "Test App"

    def ready(self):
        if getattr(settings, "DJANGO_TYPER_THROW_TEST_EXCEPTION", False):
            raise Exception("Test ready exception")

        from .management import extensions

        register_command_plugins(extensions)



import json
import os
import subprocess
import sys
from io import StringIO
from pathlib import Path

import pytest
from django.apps import apps
from django.core.management import CommandError, call_command
from django.test import TestCase, override_settings

from django_typer.management import get_command
from tests.utils import TESTS_DIR, manage_py, run_command, similarity, rich_installed


class TestGroups(TestCase):
    """
    A collection of tests that test complex grouping commands and also that
    command inheritance behaves as expected.
    """

    def test_group_call(self):
        with self.assertRaises(NotImplementedError):
            get_command("groups")()

    @pytest.mark.skip()
    def test_get_help_from_incongruent_path(self):
        """
        The directory change screws up the code coverage - it makes the omitted
        directories get included because their relative paths dont resolve in the
        coverage output for this test. VERY ANNOYING - not sure how to fix?

        https://github.com/django-commons/django-typer/issues/44
        """
        # change dir to the first dir that is not a parent
        cwd = Path(os.getcwd())
        try:
            for directory in os.listdir("/"):
                top_dir = Path(f"/{directory}")
                try:
                    cwd.relative_to(top_dir)
                except ValueError:
                    # change cwd to the first directory that is not a parent and try
                    # to invoke help from there
                    os.chdir(top_dir)
                    result = subprocess.run(
                        [
                            sys.executable,
                            manage_py.absolute(),
                            "groups",
                            "--no-color",
                            "--help",
                        ],
                        capture_output=True,
                        text=True,
                        env=os.environ,
                    )
                    self.assertGreater(
                        sim := similarity(
                            result.stdout,
                            (
                                TESTS_DIR / "apps" / "test_app" / "helps" / "groups.txt"
                            ).read_text(encoding="utf-8"),
                        ),
                        0.99,  # width inconsistences drive this number < 1
                    )
                    return
        finally:
            os.chdir(cwd)

    @pytest.mark.rich
    @pytest.mark.no_rich
    @override_settings(
        INSTALLED_APPS=[
            "tests.apps.test_app",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
    )
    def test_helps(self, app="test_app"):
        for cmds in [
            ("groups",),
            ("groups", "echo"),
            ("groups", "math"),
            ("groups", "math", "divide"),
            ("groups", "math", "multiply"),
            ("groups", "string"),
            ("groups", "string", "case"),
            ("groups", "string", "case", "lower"),
            ("groups", "string", "case", "upper"),
            ("groups", "string", "strip"),
            ("groups", "string", "split"),
            ("groups", "setting"),
            ("groups", "setting", "print"),
        ]:
            if app == "test_app" and cmds[-1] in ["strip", "setting", "print"]:
                with self.assertRaises(LookupError):
                    cmd = get_command(cmds[0], stdout=buffer, no_color=True)
                    self.assertTrue(cmd.no_color)
                    cmd.print_help("./manage.py", *cmds)
                continue

            buffer = StringIO()
            cmd = get_command(cmds[0], stdout=buffer, no_color=True)
            cmd.print_help("./manage.py", *cmds)
            hlp = buffer.getvalue()
            helps_dir = "helps" if rich_installed else "helps_no_rich"
            self.assertGreater(
                sim := similarity(
                    hlp,
                    (
                        TESTS_DIR / "apps" / app / helps_dir / f"{cmds[-1]}.txt"
                    ).read_text(encoding="utf-8"),
                ),
                0.99,  # width inconsistences drive this number < 1
            )
            print(f"{app}: {' '.join(cmds)} = {sim:.2f}")

            cmd = get_command(cmds[0], stdout=buffer, force_color=True)
            cmd.print_help("./manage.py", *cmds)
            hlp = buffer.getvalue()
            if rich_installed:
                self.assertTrue("\x1b" in hlp)

    @pytest.mark.rich
    @pytest.mark.no_rich
    @override_settings(
        INSTALLED_APPS=[
            "tests.apps.test_app2",
            "tests.apps.test_app",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
    )
    def test_helps_override(self):
        print("test_helps_override")
        self.test_helps.__wrapped__(self, app="test_app2")

    @override_settings(
        INSTALLED_APPS=[
            "tests.apps.test_app",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
    )
    def test_command_line(self, settings=None):
        from tests.apps.test_app.management.commands.groups import (
            Command as Groups,
        )

        override = settings is not None
        settings = ("--settings", settings) if settings else []

        stdout, stderr, retcode = run_command("groups", *settings, "echo", "hey!")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "hey!")

        self.assertEqual(
            call_command("groups", "echo", "hey!").strip(),
            "hey!",
        )
        self.assertEqual(
            get_command("groups", "echo")("hey!").strip(),
            "hey!",
        )
        self.assertEqual(
            get_command("groups", "echo")(message="hey!").strip(),
            "hey!",
        )

        self.assertEqual(get_command("groups", Groups).echo("hey!").strip(), "hey!")

        self.assertEqual(
            get_command("groups", Groups).echo(message="hey!").strip(), "hey!"
        )

        result = run_command("groups", "--no-color", *settings, "echo", "hey!", "5")
        if override:
            self.assertEqual(result[0].strip(), ("hey! " * 5).strip())
            self.assertEqual(
                get_command("groups", Groups).echo("hey!", 5).strip(),
                ("hey! " * 5).strip(),
            )
            self.assertEqual(
                get_command("groups", Groups).echo(message="hey!", echoes=5).strip(),
                ("hey! " * 5).strip(),
            )
            self.assertEqual(
                call_command("groups", "echo", "hey!", "5").strip(),
                ("hey! " * 5).strip(),
            )
            self.assertEqual(
                call_command("groups", "echo", "hey!", echoes=5).strip(),
                ("hey! " * 5).strip(),
            )
        else:
            self.assertIn("manage.py groups echo [OPTIONS] MESSAGE", result[0])
            self.assertIn("Got unexpected extra argument (5)", result[1])
            with self.assertRaises(TypeError):
                call_command("groups", "echo", "hey!", echoes=5)
            with self.assertRaises(TypeError):
                get_command("groups", Groups).echo(message="hey!", echoes=5)

        self.assertEqual(
            run_command(
                "groups",
                *settings,
                "math",
                "--precision",
                "5",
                "multiply",
                "1.2",
                "3.5",
                " -12.3",
                parse_json=False,
            )[0].strip(),
            "-51.66000",
        )

        grp_cmd = get_command("groups", Groups)
        grp_cmd.math(precision=5)
        self.assertEqual(grp_cmd.multiply(1.2, 3.5, [-12.3]), "-51.66000")

        self.assertEqual(
            call_command(
                "groups", "math", "multiply", "1.2", "3.5", " -12.3", precision=5
            ),
            "-51.66000",
        )

        self.assertEqual(
            call_command(
                "groups", "math", "--precision=5", "multiply", "1.2", "3.5", " -12.3"
            ),
            "-51.66000",
        )

        self.assertEqual(
            run_command(
                "groups",
                *settings,
                "math",
                "divide",
                "1.2",
                "3.5",
                " -12.3",
                parse_json=False,
            )[0].strip(),
            "-0.03",
        )

        self.assertEqual(
            call_command(
                "groups",
                "math",
                "divide",
                "1.2",
                "3.5",
                " -12.3",
            ),
            "-0.03",
        )

        self.assertEqual(
            get_command("groups", Groups).divide(1.2, 3.5, [-12.3]), "-0.03"
        )
        self.assertEqual(
            get_command("groups", "math", "divide")(1.2, 3.5, [-12.3]), "-0.03"
        )

        self.assertEqual(
            run_command("groups", *settings, "string", "ANNAmontes", "case", "lower")[
                0
            ].strip(),
            "annamontes",
        )

        self.assertEqual(
            call_command("groups", "string", "ANNAmontes", "case", "lower"),
            "annamontes",
        )

        grp_cmd = get_command("groups", Groups)
        grp_cmd.string("ANNAmontes")
        self.assertEqual(grp_cmd.lower(), "annamontes")

        self.assertEqual(
            run_command("groups", *settings, "string", "annaMONTES", "case", "upper")[
                0
            ].strip(),
            "ANNAMONTES",
        )

        grp_cmd.string("annaMONTES")
        self.assertEqual(grp_cmd.upper(), "ANNAMONTES")

        self.assertEqual(
            run_command(
                "groups",
                *settings,
                "string",
                "ANNAMONTES",
                "case",
                "lower",
                "--begin",
                "4",
                "--end",
                "9",
            )[0].strip(),
            "ANNAmonteS",
        )

        self.assertEqual(
            call_command(
                "groups",
                "string",
                "ANNAMONTES",
                "case",
                "lower",
                "--begin",
                "4",
                "--end",
                "9",
            ).strip(),
            "ANNAmonteS",
        )

        self.assertEqual(
            call_command(
                "groups", "string", "ANNAMONTES", "case", "lower", begin=4, end=9
            ).strip(),
            "ANNAmonteS",
        )

        grp_cmd.string("ANNAMONTES")
        self.assertEqual(grp_cmd.lower(begin=4, end=9), "ANNAmonteS")
        grp_cmd.string("ANNAMONTES")
        self.assertEqual(grp_cmd.lower(4, 9), "ANNAmonteS")

        result = run_command(
            "groups",
            "--no-color",
            *settings,
            "string",
            "annamontes",
            "case",
            "upper",
            "4",
            "9",
        )
        if override:
            self.assertIn(
                "manage.py groups string STRING case upper [OPTIONS]",
                result[0],
            )
            self.assertIn("Got unexpected extra arguments (4 9)", result[1].strip())
            grp_cmd.string("annamontes")
            with self.assertRaises(TypeError):
                self.assertEqual(grp_cmd.upper(4, 9), "annaMONTEs")

            with self.assertRaises(CommandError):
                self.assertEqual(
                    call_command(
                        "groups", "string", "annamontes", "case", "upper", "4", "9"
                    ).strip(),
                    "annaMONTEs",
                )
        else:
            result = result[0].strip()
            self.assertEqual(result, "annaMONTEs")
            grp_cmd.string("annamontes")
            self.assertEqual(grp_cmd.upper(4, 9), "annaMONTEs")
            self.assertEqual(
                call_command(
                    "groups", "string", "annamontes", "case", "upper", "4", "9"
                ).strip(),
                "annaMONTEs",
            )

        result = run_command(
            "groups",
            "--no-color",
            *settings,
            "string",
            " emmatc  ",
            "strip",
            parse_json=False,
        )
        if override:
            self.assertEqual(result[0].strip(), "emmatc")
            self.assertEqual(
                call_command("groups", "string", " emmatc  ", "strip"), "emmatc"
            )
            grp_cmd.string(" emmatc  ")
            self.assertEqual(grp_cmd.strip(), "emmatc")
        else:
            self.assertIn(
                "manage.py groups string [OPTIONS] STRING COMMAND [ARGS]",
                result[0],
            )
            self.assertIn("No such command 'strip'.", result[1])
            with self.assertRaises(CommandError):
                self.assertEqual(
                    call_command("groups", "string", " emmatc  ", "strip"), "emmatc"
                )
            with self.assertRaises(AttributeError):
                grp_cmd.string(" emmatc  ")
                self.assertEqual(grp_cmd.strip(), "emmatc")

        self.assertEqual(
            run_command(
                "groups", *settings, "string", "c,a,i,t,l,y,n", "split", "--sep", ","
            )[0].strip(),
            "c a i t l y n",
        )
        self.assertEqual(
            call_command(
                "groups", "string", "c,a,i,t,l,y,n", "split", "--sep", ","
            ).strip(),
            "c a i t l y n",
        )
        self.assertEqual(
            call_command("groups", "string", "c,a,i,t,l,y,n", "split", sep=",").strip(),
            "c a i t l y n",
        )
        grp_cmd.string("c,a,i,t,l,y,n")
        self.assertEqual(grp_cmd.split(sep=","), "c a i t l y n")
        grp_cmd.string("c,a,i,t,l,y,n")
        self.assertEqual(grp_cmd.split(","), "c a i t l y n")

    @override_settings(
        INSTALLED_APPS=[
            "tests.apps.test_app2",
            "tests.apps.test_app",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
    )
    def test_command_line_override(self):
        self.test_command_line.__wrapped__(self, settings="tests.settings.override")

    def test_no_subgroup_leaks(self):
        with self.assertRaises(CommandError):
            call_command("groups", "case", "--help")


class TestCallCommandArgs(TestCase):
    @override_settings(
        INSTALLED_APPS=[
            "tests.apps.completion",
            "tests.apps.test_app2",
            "tests.apps.test_app",
            "django.contrib.admin",
            "django.contrib.auth",
            "django.contrib.contenttypes",
            "django.contrib.sessions",
            "django.contrib.messages",
            "django.contrib.staticfiles",
        ],
    )
    def test_completion_args(self):
        # call_command converts all args to strings - todo - fix this? or accept it? fixing it
        # would require a monkey patch. I think accepting it and trying to allow Arguments to
        # be passed in as named parameters would be a good compromise. Users can always invoke
        # the typer commands directly using () or the functions directly.

        # if autocompletion ends up requiring a monkey patch, consider fixing it

        # turns out call_command will also turn options values into strings you've flagged them as required
        # and they're passed in as named parameters

        test_app = apps.get_app_config("test_app")
        test_app2 = apps.get_app_config("test_app2")

        out = StringIO()
        call_command(
            "completion",
            ["test_app", "test_app2"],
            stdout=out,
        )
        printed_options = json.loads(out.getvalue())
        self.assertEqual(
            printed_options,
            ["test_app", "test_app2"],
        )

        out = StringIO()
        printed_options = json.loads(get_command("completion")([test_app, test_app2]))
        self.assertEqual(
            printed_options,
            ["test_app", "test_app2"],
        )



from .base import *

INSTALLED_APPS = [
    "tests.apps.completion",
    "tests.apps.util",
    "django_typer",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
]

MEDIA_ROOT = str(BASE_DIR / "media")

STATIC_ROOT = BASE_DIR / "static"



import json

from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command


class TestOverloaded(TestCase):
    """
    Tests that overloaded group/command names work as expected.
    """

    def test_overloaded_direct(self):
        overloaded = get_command("overloaded")
        overloaded.test(1, flag=True)
        self.assertEqual(
            json.loads(overloaded.samename(5, flag=False)),
            {
                "samename": {"precision": 5, "flag": False},
                "test": {"precision": 1, "flag": True},
            },
        )

        overloaded.test(5, flag=False)
        self.assertEqual(
            json.loads(overloaded.samename(1, flag=True)),
            {
                "samename": {"precision": 1, "flag": True},
                "test": {"precision": 5, "flag": False},
            },
        )

        overloaded.test(1, flag=True)
        self.assertEqual(
            json.loads(overloaded.diffname(5, flag2=False)),
            {
                "diffname": {"precision2": 5, "flag2": False},
                "test": {"precision": 1, "flag": True},
            },
        )

        overloaded.test(5, flag=False)
        self.assertEqual(
            json.loads(overloaded.diffname(1, flag2=True)),
            {
                "diffname": {"precision2": 1, "flag2": True},
                "test": {"precision": 5, "flag": False},
            },
        )

    def test_overloaded_cli(self):
        result = run_command(
            "overloaded", "test", "--flag", "1", "samename", "5", "--no-flag"
        )[0]
        self.assertEqual(
            result,
            {
                "samename": {"precision": 5, "flag": False},
                "test": {"precision": 1, "flag": True},
            },
        )

        result = run_command(
            "overloaded", "test", "--no-flag", "5", "samename", "1", "--flag"
        )[0]
        self.assertEqual(
            result,
            {
                "samename": {"precision": 1, "flag": True},
                "test": {"precision": 5, "flag": False},
            },
        )
        result = run_command(
            "overloaded", "test", "--flag", "1", "diffname", "5", "--no-flag"
        )[0]
        self.assertEqual(
            result,
            {
                "diffname": {"precision2": 5, "flag2": False},
                "test": {"precision": 1, "flag": True},
            },
        )

        result = run_command(
            "overloaded", "test", "--no-flag", "5", "diffname", "1", "--flag"
        )[0]
        self.assertEqual(
            result,
            {
                "diffname": {"precision2": 1, "flag2": True},
                "test": {"precision": 5, "flag": False},
            },
        )

    def test_overloaded_call_command(self):
        self.assertEqual(
            json.loads(
                call_command(
                    "overloaded",
                    ["test", "--flag", "1", "samename", "5", "--no-flag"],
                )
            ),
            {
                "samename": {"precision": 5, "flag": False},
                "test": {"precision": 1, "flag": True},
            },
        )
        self.assertEqual(
            json.loads(
                call_command(
                    "overloaded",
                    ["test", "--no-flag", "5", "samename", "1", "--flag"],
                )
            ),
            {
                "test": {"precision": 5, "flag": False},
                "samename": {"precision": 1, "flag": True},
            },
        )

        ret = json.loads(
            call_command("overloaded", ["test", "5", "samename", "1"], flag=True)
        )
        self.assertEqual(
            ret,
            {
                "test": {"precision": 5, "flag": True},
                "samename": {"precision": 1, "flag": True},
            },
        )

        self.assertEqual(
            json.loads(
                call_command(
                    "overloaded",
                    ["test", "--no-flag", "5", "diffname", "1", "--flag"],
                )
            ),
            {
                "diffname": {"precision2": 1, "flag2": True},
                "test": {"precision": 5, "flag": False},
            },
        )
        self.assertEqual(
            json.loads(
                call_command(
                    "overloaded",
                    ["test", "--flag", "1", "diffname", "5", "--no-flag"],
                )
            ),
            {
                "diffname": {"precision2": 5, "flag2": False},
                "test": {"precision": 1, "flag": True},
            },
        )
        self.assertEqual(
            json.loads(
                call_command(
                    "overloaded", ["test", "5", "diffname", "1"], flag=True, flag2=False
                )
            ),
            {
                "diffname": {"precision2": 1, "flag2": False},
                "test": {"precision": 5, "flag": True},
            },
        )


class TestGroupOverloads(TestCase):
    def test_grp_overload_run(self):
        stdout, stderr, retcode = run_command("grp_overload", "g0", "l2", "1")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g0:l2(1)")

        stdout, stderr, retcode = run_command("grp_overload", "g1", "l2", "a")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g1:l2(a)")

        stdout, stderr, retcode = run_command("grp_overload", "g0", "l2", "1", "cmd")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g0:l2:cmd()")

        stdout, stderr, retcode = run_command("grp_overload", "g1", "l2", "a", "cmd")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g1:l2:cmd()")

        stdout, stderr, retcode = run_command("grp_overload", "g0", "l2", "1", "cmd2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g0:l2:cmd2()")

        stdout, stderr, retcode = run_command("grp_overload", "g1", "l2", "a", "cmd2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "g1:l2:cmd2()")

    def test_grp_overload_call(self):
        stdout = call_command("grp_overload", "g0", "l2", "1")
        self.assertEqual(stdout.strip(), "g0:l2(1)")

        stdout = call_command("grp_overload", "g1", "l2", "a")
        self.assertEqual(stdout.strip(), "g1:l2(a)")

        stdout = call_command("grp_overload", "g0", "l2", "1", "cmd")
        self.assertEqual(stdout.strip(), "g0:l2:cmd()")

        stdout = call_command("grp_overload", "g1", "l2", "a", "cmd")
        self.assertEqual(stdout.strip(), "g1:l2:cmd()")

        stdout = call_command("grp_overload", "g0", "l2", "1", "cmd2")
        self.assertEqual(stdout.strip(), "g0:l2:cmd2()")

        stdout = call_command("grp_overload", "g1", "l2", "a", "cmd2")
        self.assertEqual(stdout.strip(), "g1:l2:cmd2()")

    def test_grp_overload_direct(self):
        from tests.apps.test_app.management.commands.grp_overload import (
            Command as GrpOverload,
        )

        grp_overload = get_command("grp_overload", GrpOverload)

        print("----------")
        self.assertEqual(grp_overload.g1.l2("a"), "g1:l2(a)")
        print("----------")
        self.assertEqual(grp_overload.g0.l2(1), "g0:l2(1)")
        print("----------")
        self.assertEqual(grp_overload.g0.l2.cmd(), "g0:l2:cmd()")
        print("----------")
        self.assertEqual(grp_overload.g1.l2.cmd(), "g1:l2:cmd()")
        print("----------")

        self.assertTrue(hasattr(grp_overload.g0.l2, "cmd2"))
        self.assertEqual(grp_overload.g0.l2.cmd2(), "g0:l2:cmd2()")
        self.assertEqual(grp_overload.g1.l2.cmd2(), "g1:l2:cmd2()")



from django_typer.management import TyperCommand, COMMON_DEFAULTS, initialize, command
from typer.models import Context as TyperContext


class Command(TyperCommand):
    help = "Test that django parameter suppression works as expected"

    suppressed_base_arguments = {
        "verbosity",
        "skip_checks",
        "traceback",
        "show_locals",
        "hide_locals",
    }

    def handle(self, ctx: TyperContext, traceback: bool = True):
        assert self.suppressed_base_arguments
        assert self.__class__ is Command
        assert isinstance(ctx, TyperContext)
        set(COMMON_DEFAULTS.keys())
        assert ctx.params
        assert "verbosity" not in ctx.params
        assert "skip_checks" not in ctx.params
        assert "traceback" in ctx.params
        for param in COMMON_DEFAULTS.keys():
            if param not in self.suppressed_base_arguments:
                assert param in ctx.params
        return f"traceback={traceback}"



import typing as t
from django_typer.management import TyperCommand, DTGroup, command, group
from click import Context


class AlphabetizeCommands(DTGroup):
    def list_commands(self, ctx: Context) -> t.List[str]:
        return list(sorted(self.commands.keys()))


class Command(TyperCommand, cls=AlphabetizeCommands):
    @command()
    def b(self):
        print("b")

    @command()
    def a(self):
        print("a")

    @group(cls=AlphabetizeCommands)
    def d(self):
        print("d")

    @d.command()
    def f(self):
        print("f")

    @d.command()
    def e(self):
        print("e")

    @command()
    def c(self):
        print("c")



import typing as t
from django.db import models
from tests.apps.test_app.models import ChoicesShellCompleteTester


def get_ip_choices() -> t.List[t.Tuple[str, str]]:
    return ChoicesShellCompleteTester.IP_CHOICES


class ChoicesShellCompleteTesterDJ5Plus(models.Model):
    CHAR_CHOICES = ChoicesShellCompleteTester.CHAR_CHOICES

    char_choice = models.CharField(
        max_length=12,
        choices=CHAR_CHOICES,
        db_index=True,
        default=None,
        null=True,
    )

    INT_CHOICES = {k: v for k, v in ChoicesShellCompleteTester.INT_CHOICES}

    int_choice = models.PositiveSmallIntegerField(
        choices=INT_CHOICES,
        db_index=True,
        default=None,
        null=True,
    )

    ip_choice = models.GenericIPAddressField(
        choices=get_ip_choices,
        db_index=True,
        default=None,
        null=True,
    )



from django_typer.management import TyperCommand


class Command(TyperCommand):
    def handle(self, arg1: str, arg2: str, arg3: float = 0.5, arg4: int = 1):
        """
        A basic command that uses Typer
        """



import typing as t
from django_typer.management import TyperCommand
from typer import Option


def complete(ctx, param, incomplete):
    from path_test import mod

    return [mod.option]


class Command(TyperCommand):
    def handle(
        self,
        option: t.Annotated[
            str,
            Option(
                help="An option that requires --pythonpath to be passed for tab completion to work.",
                shell_complete=complete,
            ),
        ],
    ):
        return option



import datetime
import os
import shutil
from pathlib import Path

import pytest
from django.test import TestCase
from django.db import connection

from tests.utils import run_command

BACKUP_DIRECTORY = Path(__file__).parent / "_test_archive"


class TestBackupExample(TestCase):
    databases = {"default"}

    typer = ""

    @classmethod
    def setUpClass(cls) -> None:
        if connection.vendor != "sqlite":
            run_command("migrate")
        super().setUpClass()

    def setUp(self):
        if BACKUP_DIRECTORY.exists():
            shutil.rmtree(BACKUP_DIRECTORY)
        super().setUp()

    def tearDown(self) -> None:
        if BACKUP_DIRECTORY.exists():
            shutil.rmtree(BACKUP_DIRECTORY)
        return super().tearDown()

    def test_base_backup(self):
        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--no-color",
            "--settings",
            "tests.settings.backup",
            "list",
        )
        self.assertEqual(retcode, 0, msg=stderr)
        lines = stdout.strip().splitlines()[1:]
        self.assertEqual(len(lines), 1)
        self.assertEqual(
            lines[0].strip(),
            "database(filename={database}.json, databases=['default'])",
        )

        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--settings",
            "tests.settings.backup",
            "-o",
            str(BACKUP_DIRECTORY),
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertTrue(BACKUP_DIRECTORY.exists())
        self.assertTrue((BACKUP_DIRECTORY / "default.json").exists())
        self.assertTrue(len(os.listdir(BACKUP_DIRECTORY)) == 1)

    def test_inherit_backup(self):
        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--no-color",
            "--settings",
            "tests.settings.backup_inherit",
            "list",
        )
        self.assertEqual(retcode, 0, msg=stderr)
        lines = [line.strip() for line in stdout.strip().splitlines()[1:]]
        self.assertEqual(len(lines), 2)
        self.assertTrue(
            "database(filename={database}.json, databases=['default'])" in lines
        )
        self.assertTrue("media(filename=media.tar.gz)" in lines)

        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--settings",
            "tests.settings.backup_inherit",
            "-o",
            str(BACKUP_DIRECTORY),
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertTrue(BACKUP_DIRECTORY.exists())
        self.assertTrue((BACKUP_DIRECTORY / "default.json").exists())
        self.assertTrue((BACKUP_DIRECTORY / "media.tar.gz").exists())
        self.assertTrue(len(os.listdir(BACKUP_DIRECTORY)) == 2)

    @pytest.mark.skipif(
        connection.vendor != "sqlite",
        reason="Skipped because the database should be sqlite",
    )
    def test_extend_backup(self):
        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--no-color",
            "--settings",
            "tests.settings.backup_ext",
            "list",
        )
        self.assertEqual(retcode, 0, msg=stderr)
        lines = [line.strip() for line in stdout.strip().splitlines()[1:]]
        self.assertEqual(len(lines), 3)
        self.assertTrue("database()" in lines)
        self.assertTrue("environment(filename=requirements.txt)" in lines)
        self.assertTrue("media(filename=media.tar.gz)" in lines)

        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--settings",
            "tests.settings.backup_ext",
            "-o",
            str(BACKUP_DIRECTORY),
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertTrue(BACKUP_DIRECTORY.exists())
        self.assertTrue(
            (BACKUP_DIRECTORY / f"backup_{datetime.date.today()}.sqlite3").exists()
        )
        self.assertTrue((BACKUP_DIRECTORY / "media.tar.gz").exists())
        self.assertTrue((BACKUP_DIRECTORY / "requirements.txt").exists())
        self.assertTrue(len(os.listdir(BACKUP_DIRECTORY)) == 3)

    def test_backup_pluggy(self):
        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--no-color",
            "--settings",
            "tests.settings.backup_pluggy",
            "list",
        )
        self.assertEqual(retcode, 0, msg=stderr)
        lines = [line.strip() for line in stdout.strip().splitlines()[1:]]
        self.assertEqual(len(lines), 3)
        self.assertTrue(
            "database(filename={database}.json, databases=['default'])" in lines
        )
        self.assertTrue("media(filename=media.tar.gz)" in lines)
        self.assertTrue("files()" in lines)

        stdout, stderr, retcode = run_command(
            f"backup{self.typer}",
            "--settings",
            "tests.settings.backup_pluggy",
            "-o",
            str(BACKUP_DIRECTORY),
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertTrue(BACKUP_DIRECTORY.exists())
        self.assertTrue((BACKUP_DIRECTORY / "default.json").exists())
        self.assertTrue((BACKUP_DIRECTORY / "media.tar.gz").exists())
        self.assertTrue(len(os.listdir(BACKUP_DIRECTORY)) == 2)

        self.assertTrue("files1.tar.gz" in stdout)
        self.assertTrue("files2.zip" in stdout)


class TestBackupTyperExample(TestBackupExample):
    typer = "_typer"



import json
from io import StringIO

import django
from django.core.management import call_command
from django.test import TestCase

from django_typer.management import get_command
from tests.utils import run_command, similarity


class CallbackTests(TestCase):
    cmd_name = "callback1"

    def test_helps(self, top_level_only=False):
        buffer = StringIO()
        cmd = get_command(self.cmd_name, stdout=buffer, no_color=True)
        help_output_top, stderr = run_command(self.cmd_name, "--no-color", "--help")[
            0:2
        ]
        cmd.print_help("./manage.py", self.cmd_name)
        self.assertGreaterEqual(
            similarity(help_output_top.strip(), buffer.getvalue().strip()), 0.99
        )
        self.assertIn(f"Usage: ", help_output_top)
        self.assertIn(f"manage.py {self.cmd_name} [OPTIONS]", help_output_top)

        if not top_level_only:
            buffer.truncate(0)
            buffer.seek(0)
            callback_help = run_command(
                self.cmd_name, "--no-color", "5", self.cmd_name, "--help"
            )[0]
            cmd.print_help("./manage.py", self.cmd_name, self.cmd_name)
            self.assertGreaterEqual(similarity(callback_help, buffer.getvalue()), 0.99)
            self.assertIn(f"Usage: ", callback_help)
            self.assertIn(
                f"manage.py {self.cmd_name} P1 {self.cmd_name} [OPTIONS] ARG1 ARG2",
                callback_help,
            )

    def test_command_line(self):
        self.assertEqual(
            run_command(self.cmd_name, "5", self.cmd_name, "a1", "a2")[0],
            {
                "p1": 5,
                "flag1": False,
                "flag2": True,
                "arg1": "a1",
                "arg2": "a2",
                "arg3": 0.5,
                "arg4": 1,
            },
        )

        self.assertEqual(
            run_command(
                self.cmd_name,
                "--flag1",
                "--no-flag2",
                "6",
                self.cmd_name,
                "a1",
                "a2",
                "--arg3",
                "0.75",
                "--arg4",
                "2",
            )[0],
            {
                "p1": 6,
                "flag1": True,
                "flag2": False,
                "arg1": "a1",
                "arg2": "a2",
                "arg3": 0.75,
                "arg4": 2,
            },
        )

    def test_call_command(self, should_raise=True):
        ret = json.loads(
            call_command(
                self.cmd_name,
                *["5", self.cmd_name, "a1", "a2"],
                **{"p1": 5, "arg1": "a1", "arg2": "a2"},
            )
        )
        self.assertEqual(
            ret,
            {
                "p1": 5,
                "flag1": False,
                "flag2": True,
                "arg1": "a1",
                "arg2": "a2",
                "arg3": 0.5,
                "arg4": 1,
            },
        )

        ret = json.loads(
            call_command(
                self.cmd_name,
                *[
                    "--flag1",
                    "--no-flag2",
                    "6",
                    self.cmd_name,
                    "a1",
                    "a2",
                    "--arg3",
                    "0.75",
                    "--arg4",
                    "2",
                ],
            )
        )
        self.assertEqual(
            ret,
            {
                "p1": 6,
                "flag1": True,
                "flag2": False,
                "arg1": "a1",
                "arg2": "a2",
                "arg3": 0.75,
                "arg4": 2,
            },
        )

        # show that order matters args vs options
        interspersed = [
            lambda: call_command(
                self.cmd_name,
                *[
                    "6",
                    "--flag1",
                    "--no-flag2",
                    self.cmd_name,
                    "n1",
                    "n2",
                    "--arg3",
                    "0.2",
                    "--arg4",
                    "9",
                ],
            ),
            lambda: call_command(
                self.cmd_name,
                *[
                    "--no-flag2",
                    "6",
                    "--flag1",
                    self.cmd_name,
                    "--arg4",
                    "9",
                    "n1",
                    "n2",
                    "--arg3",
                    "0.2",
                ],
            ),
        ]
        expected = {
            "p1": 6,
            "flag1": True,
            "flag2": False,
            "arg1": "n1",
            "arg2": "n2",
            "arg3": 0.2,
            "arg4": 9,
        }
        if should_raise:
            for call_cmd in interspersed:
                if should_raise:
                    with self.assertRaises(BaseException):
                        call_cmd()
                else:
                    self.assertEqual(json.loads(call_cmd()), expected)

    def test_call_command_stdout(self):
        out = StringIO()
        call_command(
            self.cmd_name,
            [
                "--flag1",
                "--no-flag2",
                "6",
                self.cmd_name,
                "a1",
                "a2",
                "--arg3",
                "0.75",
                "--arg4",
                "2",
            ],
            stdout=out,
        )

        self.assertEqual(
            json.loads(out.getvalue()),
            {
                "p1": 6,
                "flag1": True,
                "flag2": False,
                "arg1": "a1",
                "arg2": "a2",
                "arg3": 0.75,
                "arg4": 2,
            },
        )

    def test_get_version(self):
        self.assertEqual(
            str(run_command(self.cmd_name, "--version")[0]).strip(),
            django.get_version(),
        )
        self.assertEqual(
            str(run_command(self.cmd_name, "--version", "6", self.cmd_name)[0]).strip(),
            django.get_version(),
        )

    def test_call_direct(self):
        cmd = get_command(self.cmd_name)

        self.assertEqual(
            json.loads(cmd(arg1="a1", arg2="a2", arg3=0.2)),
            {"arg1": "a1", "arg2": "a2", "arg3": 0.2, "arg4": 1},
        )


class Callback2Tests(CallbackTests):
    cmd_name = "callback2"

    def test_call_command(self):
        super().test_call_command(should_raise=False)

    def test_helps(self, top_level_only=False):
        # we only run the top level help comparison because when
        # interspersed args are allowed its impossible to get the
        # subcommand to print its help
        super().test_helps(top_level_only=True)



from django_typer.management import TyperCommand

import typing as t
from enum import Enum
from pathlib import Path

from typer import Argument, Option


class VersionEnum(str, Enum):
    VERSION1 = "1"
    VERSION1_1 = "1.1"
    VERSION2 = "2"


class Command(TyperCommand):
    help = "Override some default django options."

    def handle(
        self,
        settings: t.Annotated[
            Path, Argument(help="Override default settings argument.")
        ],
        optional_arg: t.Annotated[int, Argument(help="An optional argument.")] = 0,
        version: t.Annotated[
            t.Optional[VersionEnum],
            Option("--version", help="Override default version argument."),
        ] = None,
    ):
        assert self.__class__ is Command
        return {
            "settings": settings,
            "version": version,
            **({"optional_arg": optional_arg} if optional_arg else {}),
        }



import json

from django_typer.management import TyperCommand, command, initialize


class Command(TyperCommand):
    help = "This is a test help message"

    @initialize(epilog="This is a test callback epilog")
    def preamble(self, pre_arg: str = "pre_arg"):
        assert self.__class__ is Command
        print(f"This is a test preamble, {pre_arg}")

    @command(epilog="This is a test epilog")
    def delete(self, name: str, formal: bool = False, throw: bool = False):
        """Delete something"""
        assert self.__class__ is Command
        if throw:
            raise Exception("This is a test exception")
        print(json.dumps({"name": name, "formal": formal}))

    @command(epilog="This is a test epilog")
    def create(self, name: str, number: int, switch: bool = False):
        """This is a test create command"""
        assert self.__class__ is Command
        print(json.dumps({"name": name, "number": number, "switch": switch}))



from tests.apps.test_app.management.commands.grp_overload import (
    Command as GrpOverload,
)


# print(f'g0={hex(id(GrpOverload.g0))}, g0.l2={hex(id(GrpOverload.g0.l2))}')
@GrpOverload.g0.l2.command()
def cmd2():
    return "g0:l2:cmd2()"


# print(f'g1={hex(id(GrpOverload.g1))}, g1.l2={hex(id(GrpOverload.g1.l2))}')
@GrpOverload.g1.l2.command()
def cmd2(self):
    assert self.__class__ is GrpOverload
    return "g1:l2:cmd2()"


@GrpOverload.group(invoke_without_command=True)
def g2(self):
    return "g2()"


@g2.group()
def l2():
    return "g2:l2()"



import typing as t
from click.shell_completion import CompletionItem


def custom_fallback(args: t.List[str], incomplete: str) -> t.List[CompletionItem]:
    return [CompletionItem("custom_fallback")]


def custom_fallback_cmd_str(
    args: t.List[str], incomplete: str
) -> t.List[CompletionItem]:
    return [CompletionItem(" ".join(args) + incomplete)]



import json
from io import StringIO
import pytest

import django
from django.core.management import call_command
from django.test import TestCase

from django_typer.management import TyperCommand, get_command
from tests.utils import run_command, rich_installed
from django_typer.utils import get_current_command


class BasicTests(TestCase):
    def test_common_options_function(self):
        from django_typer.management import _common_options

        self.assertIsNone(_common_options())

    def test_command_line(self):
        stdout, stderr, retcode = run_command("basic", "a1", "a2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(
            stdout,
            {"arg1": "a1", "arg2": "a2", "arg3": 0.5, "arg4": 1},
        )

        stdout, stderr, retcode = run_command(
            "basic", "a1", "a2", "--arg3", "0.75", "--arg4", "2"
        )
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(
            stdout,
            {"arg1": "a1", "arg2": "a2", "arg3": 0.75, "arg4": 2},
        )

    def test_cmd_name(self):
        self.assertEqual(
            get_command("shellcompletion", TyperCommand)._name, "shellcompletion"
        )

    def test_call_command(self):
        out = StringIO()
        returned_options = json.loads(call_command("basic", ["a1", "a2"], stdout=out))
        self.assertEqual(
            returned_options, {"arg1": "a1", "arg2": "a2", "arg3": 0.5, "arg4": 1}
        )

    def test_call_command_stdout(self):
        out = StringIO()
        call_command("basic", ["a1", "a2"], stdout=out)
        printed_options = json.loads(out.getvalue())
        self.assertEqual(
            printed_options, {"arg1": "a1", "arg2": "a2", "arg3": 0.5, "arg4": 1}
        )

    def test_get_version(self):
        self.assertEqual(
            str(run_command("basic", "--version")[0]).strip(), django.get_version()
        )

    def test_call_direct(self):
        basic = get_command("basic")
        self.assertEqual(
            json.loads(basic.handle("a1", "a2")),
            {"arg1": "a1", "arg2": "a2", "arg3": 0.5, "arg4": 1},
        )

        from tests.apps.test_app.management.commands.basic import (
            Command as Basic,
        )

        self.assertEqual(
            json.loads(Basic()("a1", "a2", arg3=0.75, arg4=2)),
            {"arg1": "a1", "arg2": "a2", "arg3": 0.75, "arg4": 2},
        )

    def test_parser(self):
        basic_cmd = get_command("basic")
        parser = basic_cmd.create_parser("./manage.py", "basic")
        with self.assertRaises(NotImplementedError):
            parser.add_argument()

    def test_command_context(self):
        basic = get_command("basic", TyperCommand)
        multi = get_command("multi", TyperCommand)
        self.assertIsNone(get_current_command())
        with basic:
            self.assertEqual(basic, get_current_command())
            with basic:
                self.assertEqual(basic, get_current_command())
                with multi:
                    self.assertEqual(multi, get_current_command())
                self.assertEqual(basic, get_current_command())
            self.assertEqual(basic, get_current_command())
        self.assertIsNone(get_current_command())

    def test_renaming(self):
        from tests.apps.test_app.management.commands.rename import (
            Command as Rename,
        )

        stdout, stderr, retcode = run_command("rename", "default")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "handle")

        stdout, stderr, retcode = run_command("rename", "renamed")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "subcommand1")

        stdout, stderr, retcode = run_command("rename", "renamed2")
        self.assertEqual(retcode, 0, msg=stderr)
        self.assertEqual(stdout.strip(), "subcommand2")

        self.assertEqual(call_command("rename", "default"), "handle")
        self.assertEqual(call_command("rename", "renamed"), "subcommand1")
        self.assertEqual(call_command("rename", "renamed2"), "subcommand2")

        self.assertEqual(get_command("rename", TyperCommand)(), "handle")
        self.assertEqual(get_command("rename", Rename).subcommand1(), "subcommand1")
        self.assertEqual(get_command("rename", Rename).subcommand2(), "subcommand2")

    def test_get_command_make_callable(self):
        args = (1, 2, 3)
        kwargs = {"named": "test!"}
        self.assertEqual(
            get_command("base")(*args, **kwargs), f"base({args}, {kwargs})"
        )

    @pytest.mark.rich
    @pytest.mark.no_rich
    def test_cmd_help_order(self):
        buffer = StringIO()
        cmd = get_command("order", TyperCommand, stdout=buffer, no_color=True)

        cmd.print_help("./manage.py", "order")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(
                hlp.index("│ order")
                < hlp.index("│ b")
                < hlp.index("│ a")
                < hlp.index("│ d")
                < hlp.index("│ c")
            )
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(
                hlp.index(" order", cmd_idx)
                < hlp.index(" b", cmd_idx)
                < hlp.index(" a", cmd_idx)
                < hlp.index(" d", cmd_idx)
                < hlp.index(" c", cmd_idx)
            )

        buffer.seek(0)
        buffer.truncate()

        cmd.print_help("./manage.py", "order", "d")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(hlp.index("│ g") < hlp.index("│ e") < hlp.index("│ f"))
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(
                hlp.index(" g", cmd_idx)
                < hlp.index(" e", cmd_idx)
                < hlp.index(" f", cmd_idx)
            )

        cmd2 = get_command("order2", TyperCommand, stdout=buffer, no_color=True)

        buffer.seek(0)
        buffer.truncate()

        cmd2.print_help("./manage.py", "order2")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(
                hlp.index("│ order2")
                < hlp.index("│ b")
                < hlp.index("│ a ")
                < hlp.index("│ d")
                < hlp.index("│ c")
                < hlp.index("│ bb")
                < hlp.index("│ aa")
            )
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(
                hlp.index(" order2", cmd_idx)
                < hlp.index(" b", cmd_idx)
                < hlp.index(" a", cmd_idx)
                < hlp.index(" d", cmd_idx)
                < hlp.index(" c", cmd_idx)
                < hlp.index(" bb", cmd_idx)
                < hlp.index(" aa", cmd_idx)
            )

        buffer.seek(0)
        buffer.truncate()

        cmd2.print_help("./manage.py", "order2", "d")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(
                hlp.index("│ g")
                < hlp.index("│ e")
                < hlp.index("│ f")
                < hlp.index("│ i")
                < hlp.index("│ h")
                < hlp.index("│ x")
            )
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(
                hlp.index(" g", cmd_idx)
                < hlp.index(" e", cmd_idx)
                < hlp.index(" f", cmd_idx)
                < hlp.index(" i", cmd_idx)
                < hlp.index(" h", cmd_idx)
                < hlp.index(" x", cmd_idx)
            )

        buffer.seek(0)
        buffer.truncate()

        cmd2.print_help("./manage.py", "order2", "d", "x")
        hlp = buffer.getvalue()

        if rich_installed:
            self.assertTrue(hlp.index("│ z") < hlp.index("│ y"))
        else:
            cmd_idx = hlp.index("Commands")
            self.assertTrue(hlp.index(" z", cmd_idx) < hlp.index(" y", cmd_idx))

    def test_cmd_metavar(self):
        """
        Test that metavar inheritance and overrides work as expected.
        """
        from tests.apps.test_app.management.commands.metavar import (
            Command as MetavarCommand,
        )

        buffer = StringIO()
        for cmd_name in [
            "metavar",
            "metavar2",
        ]:
            cmd = get_command(cmd_name, MetavarCommand, stdout=buffer, no_color=True)

            cmd.print_help("./manage.py", cmd_name)
            self.assertIn("[CLASS OPTS]", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "cmd")
            self.assertIn("<CMD OPTS>", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "grp")
            self.assertIn("(GRP OPTS)", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "grp", "grp-cmd")
            self.assertIn("<<GRP CMD OPTS>>", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "grp", "subgrp")
            self.assertIn("(GRP OPTS)", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "grp", "subgrp", "subgrp-cmd")
            self.assertIn("(GRP OPTS)", buffer.getvalue())

            buffer.seek(0)
            buffer.truncate()

            cmd.print_help("./manage.py", cmd_name, "grp", "subgrp", "subgrp-cmd2")
            self.assertIn("(SUBGRP CMD2 OPTS)", buffer.getvalue())

        metavar_init1 = get_command(
            "metavar_init", TyperCommand, stdout=buffer, no_color=True
        )
        metavar_init1.print_help("./manage.py", "metavar_init")
        self.assertIn("[OPTIONS]", buffer.getvalue())
        buffer.seek(0)
        buffer.truncate()
        metavar_init2 = get_command(
            "metavar_init2", TyperCommand, stdout=buffer, no_color=True
        )
        metavar_init2.print_help("./manage.py", "metavar_init2")
        self.assertIn("{INIT OPTS}", buffer.getvalue())
        buffer.seek(0)
        buffer.truncate()
        metavar_init3 = get_command(
            "metavar_init3", TyperCommand, stdout=buffer, no_color=True
        )
        metavar_init3.print_help("./manage.py", "metavar_init3")
        self.assertIn("[CLASS OPTS]", buffer.getvalue())

        metavar3 = get_command("metavar3", MetavarCommand, stdout=buffer, no_color=True)
        metavar3.print_help("./manage.py", "metavar3")
        self.assertIn("{INIT OVERRIDE}", buffer.getvalue())

        buffer.seek(0)
        buffer.truncate()

        metavar3.print_help("./manage.py", "metavar3", "cmd")
        self.assertIn("(CMD OVERRIDE)", buffer.getvalue())

        buffer.seek(0)
        buffer.truncate()

        metavar3.print_help("./manage.py", "metavar3", "grp")
        self.assertIn("[GRP OVERRIDE]", buffer.getvalue())
        buffer.seek(0)
        buffer.truncate()

        metavar4 = get_command("metavar4", MetavarCommand, stdout=buffer, no_color=True)
        metavar4.print_help("./manage.py", "metavar4")
        self.assertIn("{SUBCLASS OPTS}", buffer.getvalue())



from datetime import datetime
import sys
import os
from pathlib import Path
from sphinx.ext.autodoc import between
import shutil
import django
from django.conf import settings

sys.path.append(str(Path(__file__).parent.parent.parent))
sys.path.append(str(Path(__file__).parent / 'ext'))

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'tests.settings.base')
#settings.configure()
django.setup()

import django_typer
from literalinclude import ExtendedLiteralInclude

# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
project = django_typer.__title__
copyright = django_typer.__copyright__
author = django_typer.__author__
release = django_typer.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinxcontrib_django',
    'sphinx.ext.autodoc',
    'sphinx.ext.todo',
    'sphinxcontrib.typer',
    'sphinx_tabs.tabs',
    "sphinx.ext.viewcode",
    'sphinx.ext.intersphinx'
]

svg2pdf_converter = ("cairosvg",)

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# html_theme = 'sphinx_rtd_theme'
html_theme = 'furo'
html_theme_options = {
    "source_repository": "https://github.com/django-commons/django-typer/",
    "source_branch": "main",
    "source_directory": "doc/source",
}
html_title = f"{project} {release}"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']
html_css_files = ['style.css']

todo_include_todos = True

latex_engine = "xelatex"

suppress_warnings = ['app.add_directive']

linkcheck_ignore = [
    r'https://github.com/django/django/blob/main/django/core/management/__init__.py#L278',  # Ignore exact match
]

autodoc_typehints = "description"  # or signature
autodoc_typehints_format = "short"

intersphinx_mapping = {
    "django": (
        "https://docs.djangoproject.com/en/stable",
        "https://docs.djangoproject.com/en/stable/_objects/",
    ),
    "click": ("https://click.palletsprojects.com/en/stable", None),
    "rich": ("https://rich.readthedocs.io/en/stable", None),
    "python": ('https://docs.python.org/3', None)
}


def pypi_role(name, rawtext, text, lineno, inliner, options={}, content=[]):
    from docutils import nodes

    url = f"https://pypi.org/project/{text}/"
    node = nodes.reference(rawtext, text, refuri=url, **options)
    return [node], []


def setup(app):
    from docutils.parsers.rst import roles
    # Register a sphinx.ext.autodoc.between listener to ignore everything
    # between lines that contain the word IGNORE
    app.connect(
        'autodoc-process-docstring',
        between('^.*[*]{79}.*$', exclude=True)
    )
    # app.connect('html-page-context', add_page_class)
    roles.register_local_role("pypi", pypi_role)
    app.add_directive("literalinclude", ExtendedLiteralInclude)
    app.add_crossref_type(directivename="django-admin", rolename="django-admin")
    # https://sphinxcontrib-typer.readthedocs.io/en/latest/howto.html#build-to-multiple-formats
    if Path(app.doctreedir).exists():
        shutil.rmtree(app.doctreedir)
    return app



from django_typer.management import TyperCommand, command, group, initialize


class Command(TyperCommand):
    help = "Test that inheritence extensions do not interfere with base classes."

    arg5_0: bool
    subsubgroup_called: bool = False

    @initialize(invoke_without_command=True)
    def init(self, arg: bool = True):
        assert isinstance(self, Command)
        return f"test_app::interference::init({arg})"

    @command()
    def cmd1(self, arg1_0: str):
        """
        Command 1, takes an str.
        """
        assert isinstance(self, Command)
        return f"test_app::interference::cmd1({arg1_0})"

    @command()
    def cmd2(self, arg2_0: int):
        """
        Command 2, takes an int.
        """
        assert isinstance(self, Command)
        return f"test_app::interference::cmd2({arg2_0})"

    @group(invoke_without_command=True)
    def grp1(self, argg3: float):
        """
        Group 1, take a float.
        """
        assert isinstance(self, Command)
        self.argg3 = argg3
        return f"test_app::interference::grp1({argg3})"

    @grp1.command()
    def cmd3(self, arg3_0: str, arg3_1: str):
        """
        Command 3, take two strings.
        """
        assert isinstance(self, Command)
        return f"test_app::interference::grp1({self.argg3})::cmd3({arg3_0}, {arg3_1})"

    @grp1.command()
    def cmd4(self, arg4_0: int, arg4_1: int):
        """
        Command 4, take two ints.
        """
        assert isinstance(self, Command)
        return f"test_app::interference::grp1({self.argg3})::cmd4({arg4_0}, {arg4_1})"

    @grp1.group(invoke_without_command=True)
    def subgroup(self, arg5_0: bool = False):
        assert isinstance(self, Command)
        self.arg5_0 = arg5_0
        return f"test_app::interference::grp1({self.argg3})::subgroup({arg5_0})"

    @subgroup.command()
    def subcommand(self):
        assert isinstance(self, Command)
        return f"test_app::interference::grp1({self.argg3})::subgroup({self.arg5_0})::subcommand()"

    @subgroup.group()
    def subsubgroup(self):
        assert isinstance(self, Command)
        self.subsubgroup_called = True

    @subsubgroup.command()
    def subsubcommand(self):
        assert isinstance(self, Command)
        return f"test_app::interference::grp1({self.argg3})::subgroup({self.arg5_0})::subsubgroup({self.subsubgroup_called})::subsubcommand()"



import typing as t

from django.utils.translation import gettext_lazy as _
from typer import Argument, Option

from django_typer.management import TyperCommand
from django_typer.completers.model import ModelObjectCompleter
from django_typer.parsers.model import ModelObjectParser
from tests.apps.examples.polls.models import Question as Poll


class Command(TyperCommand):
    def handle(
        self,
        polls: t.Annotated[
            t.List[Poll],
            Argument(
                parser=ModelObjectParser(Poll),
                shell_complete=ModelObjectCompleter(
                    Poll, help_field="question_text"
                ),
                help=_("The database IDs of the poll(s) to close."),
            ),
        ],
        delete: t.Annotated[
            bool,
            Option(
                "--delete",  # we can also get rid of that unnecessary --no-delete flag
                help=_("Delete poll instead of closing it."),
            ),
        ] = False,
    ):
        """
        Closes the specified poll for voting.


        As mentioned in the last section, helps can also
        be set in the docstring
        """
        for poll in polls:
            poll.opened = False
            poll.save()
            self.stdout.write(
                self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
            )
            if delete:
                poll.delete()



import os
from django.conf import settings
import typer
import json
import typing as t
import shutil

from django_typer.management import TyperCommand
from pathlib import Path
from django.core.management import call_command
from django.contrib import admindocs

django_typer = Path(__file__).parent.parent.parent.parent.parent.parent / "django_typer"
# temp_typer_link = django_typer / "typer"


class Command(TyperCommand):
    help = "(Re)Generate the translations for django-typer help strings."

    def handle(self):
        cwd = os.getcwd()
        # if temp_typer_link.is_symlink():
        #     os.unlink(temp_typer_link)
        # os.symlink(Path(typer.__file__).parent, temp_typer_link)
        # use the languages that django has admin documentation in
        languages = sorted(os.listdir(Path(admindocs.__file__).parent / "locale"))
        # lang_args = (list(itertools.chain.from_iterable(("-l", lang) for lang in languages)))
        try:
            os.chdir(django_typer)

            for lang in languages:
                locale_dir = Path(django_typer) / f"locale/{lang}"
                call_command("makemessages", "-l", lang, symlinks=True)
                try:
                    call_command("translate_messages", "-l", lang)
                    po_file = locale_dir / "LC_MESSAGES" / "django.po"
                    if po_file.is_file():
                        po = po_file.read_text()
                        if "&quot;" in po:
                            po_file.write_text(po.replace("&quot;", '\\"'))
                except Exception as e:
                    self.secho(
                        f"Failed to translate messages for {lang}: {e}", fg="red"
                    )
                    if locale_dir.is_dir():
                        shutil.rmtree(Path(django_typer) / f"locale/{lang}")
            call_command("compilemessages")
        finally:
            os.chdir(cwd)
            # if temp_typer_link.is_symlink():
            #    os.unlink(temp_typer_link)



import sys

import click
from django.core.management.base import CommandError

from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    help = "Test traceback scenarios."

    @command()
    def error(self, throw_command: bool = False, throw_other: bool = False):
        assert self.__class__ is Command
        if throw_command and throw_other:
            raise click.exceptions.UsageError(
                "--throw-command and --throw-other are mutually exclusive."
            )
        if throw_command:
            raise CommandError("This command failed!")
        if throw_other:
            raise RuntimeError("This command threw an unexpected error!")

    @command()
    def exit(self, code: int = 0):
        assert self.__class__ is Command
        sys.exit(code)



from django_typer.management import command, initialize

from .upstream import Command as Upstream


# inherit directly from the upstream command class
class Command(Upstream):
    # override init
    @initialize()
    def init(self):
        return "downstream:init"

    # override sub1
    @command()
    def sub1(self):
        return "downstream:sub1"

    # add a 3rd top level command
    @command()
    def sub3(self):
        return "downstream:sub3"

    # add a new subcommand to grp1
    @Upstream.grp1.command()
    def grp1_cmd2(self):
        return "downstream:grp1_cmd2"



from typing import Annotated

from typer import Option

from django_typer.management import TyperCommand


class Command(TyperCommand):
    help = "Test usage error behavior."

    def handle(
        self,
        arg1: int,
        flag1: bool = False,
        opt1: Annotated[int, Option(help="An option")] = 5,
    ):
        assert self.__class__ is Command



import typing as t
from datetime import date, datetime, time
from enum import Enum
from uuid import UUID

from click import Context, Parameter, ParamType
from django.core.management import CommandError
from django.db import models

from django_typer.completers.model import ModelObjectCompleter


class ReturnType(Enum):
    MODEL_INSTANCE = 0
    """Return the model instance with the matching field value."""

    FIELD_VALUE = 1
    """Return the value of the field that was matched."""

    QUERY_SET = 2
    """Return a queryset of model instances that match the field value."""


class ModelObjectParser(ParamType):
    """
    A parser that will turn strings into model object instances based on the
    configured lookup field and model class.

    .. code-block:: python

        from django_typer.parsers.model import ModelObjectParser

        class Command(TyperCommand):
            def handle(
                self,
                django_apps: Annotated[
                    t.List[MyModel],
                    typer.Argument(
                        parser=ModelObjectParser(MyModel, lookup_field="name"),
                        help=_("One or more application labels."),
                    ),
                ],
            ):

    .. note::

        Typer_ does not respect the shell_complete functions on ParamTypes passed as
        parsers. To add shell_completion see
        :class:`~django_typer.completers.ModelObjectCompleter` or the
        :func:`~django_typer.utils.model_parser_completer` convenience function.

    :param model_cls: The model class to use for lookup.
    :param lookup_field: The field to use for lookup. Defaults to 'pk'.
    :param on_error: A callable that will be called if the lookup fails.
        The callable should accept three arguments: the model class, the
        value that failed to lookup, and the exception that was raised.
        If not provided, a CommandError will be raised.
    :param return_type: The model object parser can return types other than the model
        instance (default) - use the ReturnType enumeration to return other types
        from the parser including QuerySets or the primitive values of the model fields.
    """

    error_handler = t.Callable[[t.Type[models.Model], str, Exception], None]

    model_cls: t.Type[models.Model]
    lookup_field: str
    case_insensitive: bool = False
    on_error: t.Optional[error_handler] = None
    return_type: ReturnType = ReturnType.MODEL_INSTANCE

    _lookup: str = ""
    _field: models.Field
    _completer: ModelObjectCompleter

    __name__: str = "MODEL"  # typer internals expect this

    def _get_metavar(self) -> str:
        if isinstance(self._field, models.IntegerField):
            return "INT"
        elif isinstance(self._field, models.EmailField):
            return "EMAIL"
        elif isinstance(self._field, models.URLField):
            return "URL"
        elif isinstance(self._field, models.GenericIPAddressField):
            return "[IPv4|IPv6]"
        elif isinstance(self._field, models.UUIDField):
            return "UUID"
        elif isinstance(self._field, (models.FloatField, models.DecimalField)):
            return "FLOAT"
        elif isinstance(self._field, (models.FileField, models.FilePathField)):
            return "PATH"
        elif isinstance(self._field, models.DateTimeField):
            return "ISO 8601"
        elif isinstance(self._field, models.DateField):
            return "YYYY-MM-DD"
        elif isinstance(self._field, models.TimeField):
            return "HH:MM:SS.sss"
        elif isinstance(self._field, models.DurationField):
            return "ISO 8601"
        return "TXT"

    def __init__(
        self,
        model_cls: t.Type[models.Model],
        lookup_field: t.Optional[str] = None,
        case_insensitive: bool = case_insensitive,
        on_error: t.Optional[error_handler] = on_error,
        return_type: ReturnType = return_type,
    ):
        from django.contrib.contenttypes.fields import GenericForeignKey

        self.model_cls = model_cls
        self.lookup_field = str(
            lookup_field or getattr(self.model_cls._meta.pk, "name", "id")
        )
        self.on_error = on_error
        self.return_type = return_type
        self.case_insensitive = case_insensitive
        field = self.model_cls._meta.get_field(self.lookup_field)
        assert not isinstance(field, (models.ForeignObjectRel, GenericForeignKey)), (
            "{cls} is not a supported lookup field."
        ).format(cls=self._field.__class__.__name__)
        self._field = field
        if self.case_insensitive and "iexact" in self._field.get_lookups():
            self._lookup = "__iexact"
        self.__name__ = self._get_metavar()

    def convert(
        self, value: t.Any, param: t.Optional[Parameter], ctx: t.Optional[Context]
    ):
        """
        Invoke the parsing action on the given string. If the value is
        already a model instance of the expected type the value will
        be returned. Otherwise the value will be treated as a value to query
        against the lookup_field. If no model object is found the error
        handler is invoked if one was provided.

        :param value: The value to parse.
        :param param: The parameter that the value is associated with.
        :param ctx: The context of the command.
        :raises CommandError: If the lookup fails and no error handler is
            provided.
        """
        original = value
        try:
            if not isinstance(value, str):
                return value
            elif isinstance(self._field, models.UUIDField):
                uuid = ""
                for char in value:
                    if char.isalnum():
                        uuid += char
                value = UUID(uuid)
            elif isinstance(self._field, models.DateTimeField):
                value = datetime.fromisoformat(value)
            elif isinstance(self._field, models.DateField):
                value = date.fromisoformat(value)
            elif isinstance(self._field, models.TimeField):
                value = time.fromisoformat(value)
            elif isinstance(self._field, models.DurationField):
                from django_typer.utils import parse_iso_duration

                parsed, ambiguous = parse_iso_duration(value)
                if ambiguous:
                    raise ValueError(f"Invalid duration: {value}")
                value = parsed
            if self.return_type is ReturnType.QUERY_SET:
                return self.model_cls.objects.filter(
                    **{f"{self.lookup_field}{self._lookup}": value}
                )
            elif self.return_type is ReturnType.FIELD_VALUE:
                return value
            return self.model_cls.objects.get(
                **{f"{self.lookup_field}{self._lookup}": value}
            )
        except ValueError as err:
            if self.on_error:
                return self.on_error(self.model_cls, original, err)
            raise CommandError(
                f"{original} is not a valid {self._field.__class__.__name__}"
            ) from err
        except self.model_cls.DoesNotExist as err:
            if self.on_error:
                return self.on_error(self.model_cls, original, err)
            raise CommandError(
                f"{self.model_cls.__name__}.{self.lookup_field}='{original}' does not "
                "exist!"
            ) from err



import typer

from django_typer.management import TyperCommand


class Command(TyperCommand):
    def handle(self):
        self.echo("echo does not support styling")
        self.secho("but secho does!", fg=typer.colors.GREEN)



import typing as t
from functools import cached_property
from pathlib import Path

from click.shell_completion import CompletionItem

from . import DjangoTyperShellCompleter


class BashComplete(DjangoTyperShellCompleter):
    """
    This completer class supports the bash_ shell. Completion scripts are
    installed in the ``~/.bash_completions`` directory and are sourced in
    ``~/.bashrc``.

    Bash_ does not support help text in completions so these are not returned.
    The format of the completion is ``type,value`` and different suggestions
    are separated by newlines.

    See also: https://github.com/scop/bash-completion#faq
    """

    name = "bash"
    """
    shell executable.
    """

    template = "shell_complete/bash.sh"
    """
    The template used to render the bash completion script.
    """

    supports_scripts = True
    """
    The bash completer supports script invocations.
    """

    color = False
    """
    Zsh_ does support ansi control codes in completion suggestions, but we disable them
    by default.
    """

    def get_user_profile(self) -> Path:
        """
        Get .bashrc it is always located in the user's home directory.
        """
        return Path.home() / ".bashrc"

    @cached_property
    def install_dir(self) -> Path:
        install_dir = self.get_user_profile().parent / ".bash_completions"
        install_dir.mkdir(parents=True, exist_ok=True)
        return install_dir

    @staticmethod
    def _check_version() -> t.Optional[t.Tuple[int, int]]:
        import re
        import subprocess

        output = subprocess.run(
            ["bash", "-c", 'echo "${BASH_VERSION}"'], stdout=subprocess.PIPE
        )
        match = re.search(r"^(\d+)\.(\d+)\.\d+", output.stdout.decode())
        return (int(match.groups()[0]), int(match.groups()[1])) if match else None

    def source_vars(self) -> t.Dict[str, t.Any]:
        """
        When the bash version is 4.4 or higher, the ``nosort`` option is
        available and can be used to disable sorting of completions. We
        add an additional context variable to the template to include this:

        * ``complete_opts``: The options to pass to bash's ``complete`` command.
        """
        complete_opts = ""
        version = self._check_version()
        if version and version >= (4, 4):
            complete_opts = "-o nosort"
        return {
            **super().source_vars(),
            "complete_opts": complete_opts,
            "bash_version": version,
            "use_compopt": version and version >= (4, 4),
        }

    def format_completion(self, item: CompletionItem) -> str:
        return f"{item.type},{item.value}"

    def install(self, prompt: bool = True) -> t.List[Path]:
        assert self.prog_name
        edited = []
        script = self.install_dir / f"{self.prog_name}.sh"
        bashrc = self.get_user_profile()
        bashrc_source = bashrc.read_text() if bashrc.is_file() else ""
        source_line = f"source {script}"
        if source_line not in bashrc_source:
            bashrc_source += f"\n{source_line}\n"
            if self.prompt(
                prompt=prompt,
                source=source_line,
                file=bashrc,
                start_line=bashrc_source.count("\n") + 1,
            ):
                Path.home().mkdir(parents=True, exist_ok=True)
                with open(bashrc, "a") as f:
                    f.write(f"\n{source_line}\n")
                edited.append(bashrc)

        source = self.source()
        if self.prompt(prompt=prompt, source=source, file=script):
            script.parent.mkdir(parents=True, exist_ok=True)
            script.write_text(self.source())
            edited.append(script)
        return edited

    def uninstall(self):
        assert self.prog_name
        script = self.install_dir / f"{self.prog_name}.sh"
        if script.is_file():
            script.unlink()

        bashrc = self.get_user_profile()
        if bashrc.is_file():
            bashrc_source = bashrc.read_text()
            bashrc.write_text(bashrc_source.replace(f"source {script}\n", ""))



from django_typer.management import Typer

# here we pass chain=True to typer telling it to allow invocation of
# multiple subcommands
app = Typer(chain=True)


@app.command()
def cmd1(self):
    self.stdout.write("cmd1")


@app.command()
def cmd2(self):
    self.stdout.write("cmd2")



import os
import shutil
import sys
from pathlib import Path
import typing as t

import pytest
from django.test import TestCase

from tests.shellcompletion import (
    _ScriptCompleteTestCase,
    _InstalledScriptCompleteTestCase,
)


@pytest.mark.skipif(shutil.which("bash") is None, reason="Bash not available")
class BashTests(_ScriptCompleteTestCase, TestCase):
    shell = "bash"
    directory = Path("~/.bash_completions").expanduser()
    interactive_opt = "-i"
    tabs = "\t\t"

    environment = [
        "export DJANGO_SETTINGS_MODULE=tests.settings.completion",
        f"export PATH={Path(sys.executable).parent}:$PATH",
        "source ~/.bashrc",
        f"source {Path(sys.executable).absolute().parent / 'activate'}",
    ]

    def verify_install(self, script=None, directory: t.Optional[Path] = None):
        directory = directory or self.directory
        if not script:
            script = self.manage_script
        self.assertTrue((directory / f"{script}.sh").exists())

    def verify_remove(self, script=None, directory: t.Optional[Path] = None):
        directory = directory or self.directory
        if not script:
            script = self.manage_script
        self.assertFalse((directory / f"{script}.sh").exists())

    @pytest.mark.rich
    @pytest.mark.no_rich
    @pytest.mark.skip("Bash completer does not support rich output.")
    def test_rich_output(self): ...


@pytest.mark.skipif(shutil.which("bash") is None, reason="Bash not available")
class BashExeTests(_InstalledScriptCompleteTestCase, BashTests):
    shell = "bash"

    @pytest.mark.skipif(
        not bool(os.environ.get("ENABLE_CI_ONLY_TESTS", False)),
        reason="This test is dangerous to run on a user machine, "
        "because it may nuke their shell profile file.",
    )
    def test_no_bashrc_file(self):
        bashrc = ""
        try:
            if (Path.home() / ".bashrc").exists():
                bashrc = (Path.home() / ".bashrc").read_text()
                os.unlink(Path.home() / ".bashrc")
            self.test_shell_complete()
            os.unlink(Path.home() / ".bashrc")
            self.remove()
            self.verify_remove()
        finally:
            if bashrc:
                (Path.home() / ".bashrc").write_text(bashrc)

    def test_source_template(self):
        self.assertEqual(
            (
                Path(__file__).parent.parent.parent
                / "src/django_typer/templates/shell_complete/bash.sh"
            ).read_text(),
            self.get_completer().source_template,
        )

        self.assertEqual(
            (
                Path(__file__).parent.parent.parent
                / "src/django_typer/templates/shell_complete/zsh.sh"
            ).read_text(),
            self.get_completer(template="shell_complete/zsh.sh").source_template,
        )



import typing as t

from django.core.management.base import CommandError

from django_typer.management import Typer
from tests.apps.examples.polls.models import Question as Poll

app = Typer(help="Closes the specified poll for voting")


@app.command()
def handle(
    self,
    poll_ids: t.List[int],
    delete: bool = False,
):
    for poll_id in poll_ids:
        try:
            poll = Poll.objects.get(pk=poll_id)
        except Poll.DoesNotExist:
            raise CommandError(f'Poll "{poll_id}" does not exist')

        poll.opened = False
        poll.save()

        self.stdout.write(
            self.style.SUCCESS(f'Successfully closed poll "{poll.id}"')
        )

        if delete:
            poll.delete()



from django_typer.management import TyperCommand, command, group, initialize


class Command(TyperCommand):
    @initialize()
    def init(self):
        return "upstream:init"

    @command()
    def sub1(self):
        return "upstream:sub1"

    @command()
    def sub2(self):
        return "upstream:sub2"

    @group()
    def grp1(self):
        return "upstream:grp1"

    @grp1.command()
    def grp1_cmd1(self):
        return "upstream:grp1_cmd1"



from django.utils.translation import gettext_lazy as _

from django_typer.management import Typer, TyperCommand

from . import native_groups

Command: TyperCommand

app = Typer(native_groups.app)


@app.init_grp1.group(name="subgrp")
def run_subgrp(self, option1: bool = False, option2: bool = False):
    """
    Override SUBGROUP
    """
    assert not isinstance(self, native_groups.Command)
    assert isinstance(self, Command)
    self.option1 = option1
    self.option2 = option2


@run_subgrp.command()
def sg_cmd1(self):
    """
    Subgroup command 1. No args.
    """
    assert not isinstance(self, native_groups.Command)
    assert isinstance(self, Command)
    return {
        "verbosity": native_groups._verbosity,
        "flag": native_groups._flag,
        "option1": self.option1,
        "option2": self.option2,
    }


@run_subgrp.command(help=_("Subgroup command 2, Takes an int."))
def sg_cmd2(self, number: int):
    assert not isinstance(self, native_groups.Command)
    assert isinstance(self, Command)
    return {
        "verbosity": native_groups._verbosity,
        "flag": native_groups._flag,
        "option1": self.option1,
        "option2": self.option2,
        "number": number,
    }



from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    help = "Test various forms of handle override."

    @command()
    def handle(self) -> str:
        assert self.__class__ is Command
        return "handle2"



import json
import typing as t
from pathlib import Path
from typing import Annotated
from functools import partial

import typer
from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

from django_typer.management import TyperCommand
from django_typer.completers.apps import app_labels
from django_typer.parsers.apps import app_config
from django_typer.completers.path import (
    directories,
    paths,
    import_paths,
    static_paths,
    media_paths,
)
from django_typer.completers.settings import setting, languages
from django_typer.completers.db import databases
from django_typer.completers.cmd import commands
from django_typer.completers import these_strings, chain
from django_typer.types import COMMON_PANEL


settings_dir = Path(__file__).parent.parent.parent.parent.parent / "settings"


class Command(TyperCommand, rich_markup_mode="rich"):
    def handle(
        self,
        django_apps: Annotated[
            t.List[AppConfig],
            typer.Argument(
                help=t.cast(str, _("One or more application labels.")),
                parser=app_config,
                shell_complete=app_labels,
            ),
        ],
        option: Annotated[
            t.Optional[AppConfig],
            typer.Option(
                parser=app_config,
                help=t.cast(str, _("An app given as an option.")),
                shell_complete=app_labels,
            ),
        ] = None,
        path: Annotated[
            t.Optional[Path],
            typer.Option(
                help=t.cast(str, _("A path given as an option.")),
                shell_complete=paths,
            ),
        ] = None,
        dir: Annotated[
            t.Optional[Path],
            typer.Option(
                help=t.cast(str, _("A directory given as an option.")),
                shell_complete=directories,
            ),
        ] = None,
        strings_unique: Annotated[
            t.Optional[t.List[str]],
            typer.Option(
                "--str",
                help=t.cast(str, _("A list of unique strings.")),
                shell_complete=these_strings(["str1", "str2", "ustr"]),
            ),
        ] = None,
        strings_duplicates: Annotated[
            t.Optional[t.List[str]],
            typer.Option(
                "--dup",
                help=t.cast(str, _("A list of strings that can have duplicates.")),
                shell_complete=these_strings(
                    ["str1", "str2", "ustr"], allow_duplicates=True
                ),
            ),
        ] = None,
        commands: Annotated[
            t.List[str],
            typer.Option(
                "--cmd",
                help=t.cast(
                    str,
                    _("A command by [bold]import path[/bold] or [bold]name[/bold]."),
                ),
                shell_complete=chain(import_paths, commands()),
            ),
        ] = [],
        command_dups: Annotated[
            t.List[str],
            typer.Option(
                "--cmd-dup",
                help=t.cast(
                    str,
                    _("A list of [reverse]commands[/reverse] by import path or name."),
                ),
                shell_complete=chain(
                    import_paths, commands(allow_duplicates=True), allow_duplicates=True
                ),
            ),
        ] = [],
        command_first: Annotated[
            t.List[str],
            typer.Option(
                "--cmd-first",
                help=t.cast(
                    str,
                    _(
                        "A list of [yellow][underline]commands[/underline][/yellow] by either import path or name."
                    ),
                ),
                shell_complete=chain(
                    import_paths, commands(allow_duplicates=True), first_match=True
                ),
            ),
        ] = [],
        app_opt: Annotated[
            t.List[str],
            typer.Option(
                help=t.cast(str, _("One or more application labels.")),
                shell_complete=app_labels,
            ),
        ] = ["test_app"],
        databases: Annotated[
            t.List[str],
            typer.Option(
                "--db",
                help=t.cast(str, _("One or more database aliases.")),
                shell_complete=databases(),
            ),
        ] = [],
        statics: Annotated[
            t.Optional[Path],
            typer.Option(
                help=t.cast(str, _("Static files.")),
                shell_complete=static_paths,
            ),
        ] = None,
        media: Annotated[
            t.Optional[Path],
            typer.Option(
                help=t.cast(str, _("Media files.")),
                shell_complete=media_paths,
            ),
        ] = None,
        settings_module: Annotated[
            str,
            typer.Option(
                help=t.cast(
                    str,
                    ("The python path to the settings file under tests.settings."),
                ),
                rich_help_panel=COMMON_PANEL,
                shell_complete=partial(import_paths, root=settings_dir),
                show_default=False,
            ),
        ] = "",
        languages: Annotated[
            t.List[str],
            typer.Option(
                "--lang",
                help=t.cast(str, _("One or more languages.")),
                shell_complete=languages,
            ),
        ] = [],
        setting: Annotated[
            t.List[str],
            typer.Option(
                "--setting",
                help=t.cast(str, _("One or more Django setting variables.")),
                shell_complete=setting,
            ),
        ] = [],
    ):
        assert self.__class__ is Command
        for app in django_apps:
            assert isinstance(app, AppConfig)
        if option:
            return json.dumps(
                {
                    "django_apps": [app.label for app in django_apps],
                    "option": option.label,
                }
            )
        return json.dumps([app.label for app in django_apps])



"""
Test options metavar propagation and overrides
"""

from django_typer.management import TyperCommand, command, group


class Command(
    TyperCommand, options_metavar="[CLASS OPTS]", rich_markup_mode=None
):  # <-- default
    # @initialize(options_metavar="{INIT OPTS}")  # <-- override
    # def init(self):
    #     print("init")

    @command(options_metavar="<CMD OPTS>")  # <-- override
    def cmd(self):
        print("cmd")

    @group(options_metavar="(GRP OPTS)")  # <-- override
    def grp(self):
        print("grp")

    @grp.command(options_metavar="<<GRP CMD OPTS>>")  # <-- override
    def grp_cmd(self):
        print("grp cmd")

    @grp.group()  # <-- inherit from grp
    def subgrp(self):
        print("subgrp")

    @subgrp.command()  # <-- inherit from grp
    def subgrp_cmd(self):
        print("subgrp cmd")

    @subgrp.command(options_metavar="(SUBGRP CMD2 OPTS)")  # <-- override
    def subgrp_cmd2(self):
        print("subgrp cmd2")



import tarfile
import typing as t
from pathlib import Path

import typer
from django.conf import settings

from django_typer.management import command
from tests.apps.examples.plugins.backup.management.commands.backup import (
    Command as Backup,
)


class Command(Backup):  # inherit from the original command
    # add a new command called media that archives the MEDIA_ROOT dir
    @command()
    def media(
        self,
        filename: t.Annotated[
            str,
            typer.Option(
                "-f",
                "--filename",
                help=("The name of the file to use for the media backup tar."),
            ),
        ] = "media.tar.gz",
    ):
        """
        Backup the media files (i.e. those files in MEDIA_ROOT).
        """
        media_root = Path(settings.MEDIA_ROOT)
        output_file = self.output_directory / filename

        # backup the media directory into the output file as a gzipped tar
        typer.echo(f"Backing up {media_root} to {output_file}")
        with tarfile.open(output_file, "w:gz") as tar:
            tar.add(media_root, arcname=media_root.name)



from django.db import models
from django.utils import timezone


class Question(models.Model):
    question_text = models.CharField(max_length=200)
    pub_date = models.DateTimeField("date published", default=timezone.now)
    opened = models.BooleanField(default=True)

    def __str__(self):
        return self.question_text

    class Meta:
        verbose_name = "Poll"
        verbose_name_plural = "Polls"


class Choice(models.Model):
    question = models.ForeignKey(Question, on_delete=models.CASCADE)
    choice_text = models.CharField(max_length=200)
    votes = models.IntegerField(default=0)

    def __str__(self):
        return self.choice_text



from django_typer.management import TyperCommand, command


class Command(TyperCommand):
    @command(name="subcommand1")
    def handle(self):
        return "handle"

    @command()
    def subcommand2(self):
        return "subcommand2"

    @command()
    def subcommand3(self):
        return "subcommand3"
